import random
import numpy as np
import copy

from detpy.models.member import Member
from detpy.models.population import Population
from detpy.models.enums.optimization import OptimizationType
def mutation_ind(base_member: Member, member1: Member, member2: Member, fs):
    """
        Formula: v_ij = x_r1 + FS(x_r2 - x_r3)
    """
    new_member = copy.deepcopy(base_member)
    new_member.chromosomes = base_member.chromosomes + (member1.chromosomes - member2.chromosomes) * fs
    return new_member


def mutation(population: Population, fs):
    new_members = []
    for _ in range(population.size):
        selected_members = random.sample(population.members.tolist(), 3)
        new_member = mutation_ind(selected_members[0], selected_members[1], selected_members[2], fs)
        new_members.append(new_member)

    new_population = Population(
        lb=population.lb,
        ub=population.ub,
        arg_num=population.arg_num,
        size=population.size,
        optimization=population.optimization
    )
    new_population.members = np.array(new_members)
    return new_population


def binomial_crossing_ind(org_member: Member, mut_member: Member, cr):
    new_member = copy.deepcopy(org_member)

    random_numbers = np.random.rand(new_member.args_num)
    mask = random_numbers <= cr
    i_rand = np.random.randint(low=0, high=new_member.args_num)

    for i in range(new_member.args_num):
        if mask[i] or i_rand == i:
            new_member.chromosomes[i].real_value = mut_member.chromosomes[i].real_value
        else:
            new_member.chromosomes[i].real_value = org_member.chromosomes[i].real_value

    return new_member


def binomial_crossing(origin_population: Population, mutated_population: Population, cr):
    if origin_population.size != mutated_population.size:
        print("Binomial_crossing: populations have different sizes")
        return None

    new_members = []
    for i in range(origin_population.size):
        new_member = binomial_crossing_ind(origin_population.members[i], mutated_population.members[i], cr)
        new_members.append(new_member)

    new_population = Population(
        lb=origin_population.lb,
        ub=origin_population.ub,
        arg_num=origin_population.arg_num,
        size=origin_population.size,
        optimization=origin_population.optimization
    )
    new_population.members = np.array(new_members)
    return new_population


def selection(origin_population: Population, modified_population: Population):
    if origin_population.size != modified_population.size:
        print("Selection: populations have different sizes")
        return None

    if origin_population.optimization != modified_population.optimization:
        print("Selection: populations have different optimization types")
        return None

    optimization = origin_population.optimization
    new_members = []
    for i in range(origin_population.size):
        if optimization == OptimizationType.MINIMIZATION:
            if origin_population.members[i] <= modified_population.members[i]:
                new_members.append(copy.deepcopy(origin_population.members[i]))
            else:
                new_members.append(copy.deepcopy(modified_population.members[i]))
        elif optimization == OptimizationType.MAXIMIZATION:
            if origin_population.members[i] >= modified_population.members[i]:
                new_members.append(copy.deepcopy(origin_population.members[i]))
            else:
                new_members.append(copy.deepcopy(modified_population.members[i]))

    new_population = Population(
        lb=origin_population.lb,
        ub=origin_population.ub,
        arg_num=origin_population.arg_num,
        size=origin_population.size,
        optimization=origin_population.optimization
    )
    new_population.members = np.array(new_members)
    return new_population


import json
import typing

from detpy.helpers.metric_helper import Metric
from detpy.models.member import Member


def get_table_name(func_name, alg_name, nr_of_args, pop_size):
    table_name = f"{func_name}_{alg_name}_" \
                 f"args{nr_of_args}_pop{pop_size}_" \
                 f"results"

    return table_name


def format_individuals(individuals: typing.List[Metric]):
    formatted_individuals = []
    for data in individuals:
        epoch = data.epoch
        best_member: Member = data.best_individual
        worst_member: Member = data.worst_individual
        mean = data.population_mean
        std = data.population_std
        exec_time = data.execution_time
        population = [str(member) for member in data.population]

        formatted_individuals.append(
            (
                epoch,
                json.dumps([chromosome.real_value for chromosome in best_member.chromosomes]),
                best_member.fitness_value,
                json.dumps([chromosome.real_value for chromosome in worst_member.chromosomes]),
                worst_member.fitness_value,
                mean,
                std,
                exec_time,
                json.dumps(population)
            )
        )
    return formatted_individuals



import numpy as np
from detpy.DETAlgs.base import BaseAlg
from detpy.DETAlgs.data.alg_data import NMDEData
from detpy.DETAlgs.methods.methods_nmde import nmde_mutation, nmde_selection, nmde_calculate_fm_crm, \
    nmde_binomial_crossing, nmde_update_f_cr
from detpy.models.enums.boundary_constrain import fix_boundary_constraints


class NMDE(BaseAlg):
    """
        NMDE

        Links:
        https://www.sciencedirect.com/science/article/pii/S0898122111000460

        References:
        Dexuan Zou, Haikuan Liu, Liqun Gao, and Steven Li. 2011. A novel modified differential evolution algorithm for
        constrained optimization problems. Comput. Math. Appl. 61, 6 (March, 2011), 1608–1623.
        https://doi.org/10.1016/j.camwa.2011.01.029
    """

    def __init__(self, params: NMDEData, db_conn=None, db_auto_write=False):
        super().__init__(NMDE.__name__, params, db_conn, db_auto_write)

        self.delta_f = params.delta_f
        self.delta_cr = params.delta_cr
        self.sp = params.sp
        self._flags = np.zeros(self.population_size)
        self._f_arr = np.random.uniform(size=self.population_size, low=0, high=2)
        self._cr_arr = np.random.uniform(size=self.population_size, low=0, high=1)
        self._f_set = set()
        self._cr_set = set()

    def next_epoch(self):
        delta_f, delta_cr, sp, flags, f_arr, cr_arr, f_set, cr_set = (
            self.delta_f, self.delta_cr, self.sp, self._flags,
            self._f_arr, self._cr_arr, self._f_set, self._cr_set
        )

        # New population after mutation
        v_pop = nmde_mutation(self._pop, f_arr)

        # Apply boundary constrains on population in place
        fix_boundary_constraints(v_pop, self.boundary_constraints_fun)

        # New population after crossing
        u_pop = nmde_binomial_crossing(self._pop, v_pop, cr_arr)

        # Update values before selection
        u_pop.update_fitness_values(self._function.eval, self.parallel_processing)

        # Select new population
        new_pop, better_members_indexes = nmde_selection(self._pop, u_pop)

        for i in range(self.population_size):
            if i in better_members_indexes:
                f_set.add(f_arr[i])
                cr_set.add(cr_arr[i])
            else:
                flags[i] += 1

        f_m, cr_m = nmde_calculate_fm_crm(f_set, cr_set)

        for i in range(self.population_size):
            if flags[i] == sp:
                if f_set != set() and cr_set != set():
                    f_arr[i], cr_arr[i] = nmde_update_f_cr(f_m, cr_m, delta_f, delta_cr)
                else:
                    f_arr[i] = np.random.uniform(low=0, high=2)
                    cr_arr[i] = np.random.uniform(low=0, high=1)
                flags[i] = 0

        # Override data
        self._pop = new_pop
        self.delta_f = delta_f
        self.delta_cr = delta_cr
        self.sp = sp
        self._flags = flags
        self._f_arr = f_arr
        self._cr_arr = cr_arr
        self._f_set = set()
        self._cr_set = set()

        self._epoch_number += 1



from random import random
import numpy as np

from detpy.models.member import Member
from detpy.models.population import Population


def calculate_central_points(population: Population) -> np.ndarray[float]:
    all_chromosomes = np.array(
        [[chromosome.real_value for chromosome in member.chromosomes] for member in population.members])

    min_values = np.min(all_chromosomes, axis=0)
    max_values = np.max(all_chromosomes, axis=0)

    return min_values + max_values


def calculate_opposite_pop(population: Population, is_initial_pop: bool) -> Population:
    central_points = np.full((population.arg_num,), population.lb + population.ub) if is_initial_pop else (
        calculate_central_points(population))

    opposite_members = []
    for member in population.members:
        opposite_member = Member(member.lb, member.ub, member.args_num)
        for i in range(member.args_num):
            opposite_member.chromosomes[i] = (member.chromosomes[i] - central_points[i]) * -1

        opposite_members.append(opposite_member)

    opposite_population = Population(
        lb=population.lb,
        ub=population.ub,
        arg_num=population.arg_num,
        size=population.size,
        optimization=population.optimization
    )
    opposite_population.members = np.array(opposite_members)

    return opposite_population


def opp_based_keep_best_individuals(population: Population, func, is_initial_pop: bool = False):
    opposite_population = calculate_opposite_pop(population, is_initial_pop)
    opposite_population.update_fitness_values(func)

    pops = np.concatenate((population.members, opposite_population.members))
    sorted_pops_indices = np.argsort([member.fitness_value for member in pops])
    sorted_pops = pops[sorted_pops_indices]
    population.members = sorted_pops[:population.size]


def opp_based_generation_jumping(population: Population, jumping_rate: float, func) -> bool:
    if random() < jumping_rate:
        opp_based_keep_best_individuals(population, func)
        return True
    return False



from detpy.DETAlgs.base import BaseAlg
from detpy.DETAlgs.data.alg_data import DELBData
from detpy.DETAlgs.methods.methods_delb import delb_mutation, delb_selection
from detpy.DETAlgs.methods.methods_de import binomial_crossing
from detpy.models.enums.boundary_constrain import fix_boundary_constraints


class DELB(BaseAlg):
    """
    DELB

    Links:
    https://www.sciencedirect.com/science/article/pii/S037722170500281X

    References:
    Kaelo, P. & Ali, Montaz. (2006). A numerical study of some modified differential evolution algorithms.
    European Journal of Operational Research. 169. 1176-1184. 10.1016/j.ejor.2004.08.047.
    """

    def __init__(self, params: DELBData, db_conn=None, db_auto_write=False):
        super().__init__(DELB.__name__, params, db_conn, db_auto_write)

        self.crossover_rate = params.crossover_rate  # Cr
        self.w_factor = params.w_factor  # w

    def next_epoch(self):
        # New population after mutation
        v_pop = delb_mutation(self._pop)

        # Apply boundary constrains on population in place
        fix_boundary_constraints(v_pop, self.boundary_constraints_fun)

        # New population after crossing
        u_pop = binomial_crossing(self._pop, v_pop, cr=self.crossover_rate)

        # Update values before selection
        u_pop.update_fitness_values(self._function.eval, self.parallel_processing)

        # Select new population
        new_pop = delb_selection(self._pop, u_pop, self.w_factor, self._function.eval)

        # Override data
        self._pop = new_pop

        self._epoch_number += 1



import numpy as np
import random
import copy
from detpy.models.enums.mutation import mutation_rand_1
from detpy.models.enums.optimization import OptimizationType
from detpy.models.population import Population
from detpy.DETAlgs.methods.methods_de import binomial_crossing_ind

def mutation(population: Population, mutation_factors: list[list[float, bool]]) -> Population:
    members = population.members.tolist()
    drew_members = [random.sample(members, 3) for _ in range(population.size)]

    new_members = [
        mutation_rand_1(selected_members, f)
        for selected_members, (f, _) in zip(drew_members, mutation_factors)
    ]

    new_population = Population(
        lb=population.lb,
        ub=population.ub,
        arg_num=population.arg_num,
        size=population.size,
        optimization=population.optimization
    )
    new_population.members = np.array(new_members)
    return new_population


def crossing(origin_population: Population, mutated_population: Population,
             crossover_rates: list[list[float, bool]]) -> Population:
    if origin_population.size != mutated_population.size:
        raise ValueError("Populations must have the same size for crossing")

    new_members = [
        binomial_crossing_ind(origin_population.members[i], mutated_population.members[i], crossover_rates[i][0])
        for i in range(origin_population.size)
    ]

    new_population = Population(
        lb=origin_population.lb,
        ub=origin_population.ub,
        arg_num=origin_population.arg_num,
        size=origin_population.size,
        optimization=origin_population.optimization
    )
    new_population.members = np.array(new_members)
    return new_population


def selection(origin_population: Population, modified_population: Population, mutation_factors: list[list[float, bool]],
              crossover_rates: list[list[float, bool]]) -> Population:
    if origin_population.size != modified_population.size:
        raise ValueError("Populations must have the same size for selection")

    if origin_population.optimization != modified_population.optimization:
        raise ValueError("Populations must have the same optimization type")

    optimization = origin_population.optimization
    new_members = []

    for i in range(origin_population.size):
        if optimization == OptimizationType.MINIMIZATION:
            if origin_population.members[i] <= modified_population.members[i]:
                new_members.append(copy.deepcopy(origin_population.members[i]))
                mutation_factors[i] = update_mutation_factor(origin_population.members[i],
                                                             modified_population.members[i], mutation_factors[i])
                crossover_rates[i] = update_crossover_rate(origin_population.members[i], modified_population.members[i],
                                                           crossover_rates[i])
            else:
                new_members.append(copy.deepcopy(modified_population.members[i]))
        else:  # MAXIMIZATION
            if origin_population.members[i] >= modified_population.members[i]:
                new_members.append(copy.deepcopy(origin_population.members[i]))
                mutation_factors[i] = update_mutation_factor(origin_population.members[i],
                                                             modified_population.members[i], mutation_factors[i])
                crossover_rates[i] = update_crossover_rate(origin_population.members[i], modified_population.members[i],
                                                           crossover_rates[i])
            else:
                new_members.append(copy.deepcopy(modified_population.members[i]))

    new_population = Population(
        lb=origin_population.lb,
        ub=origin_population.ub,
        arg_num=origin_population.arg_num,
        size=origin_population.size,
        optimization=origin_population.optimization
    )
    new_population.members = np.array(new_members)
    return new_population


def adapt_parameters(population: Population, mutation_factors: list[list[float, bool]],
                     crossover_rates: list[list[float, bool]]):
    sorted_members = population.get_best_members(population.size)
    max_fitness_value = sorted_members[-1].fitness_value
    min_fitness_value = sorted_members[0].fitness_value

    for i in range(population.size):
        if mutation_factors[i][1]:
            mutation_factors[i][1] = False
        else:
            mutation_factors[i] = [random.random() * ((max_fitness_value - min_fitness_value) / max_fitness_value),
                                   False]

        if crossover_rates[i][1]:
            crossover_rates[i][1] = False
        else:
            crossover_rates[i] = [random.random() * ((max_fitness_value - min_fitness_value) / max_fitness_value),
                                  False]


def update_mutation_factor(orig_member, mod_member, mutation_factor):
    try:
        delta = (mod_member.fitness_value - orig_member.fitness_value) / mod_member.fitness_value * random.random()
        return [mutation_factor[0] - delta, True]
    except ZeroDivisionError:
        return [0, True]


def update_crossover_rate(orig_member, mod_member, crossover_rate):
    try:
        delta = (mod_member.fitness_value - orig_member.fitness_value) / mod_member.fitness_value * random.random()
        return [crossover_rate[0] - delta, True]
    except ZeroDivisionError:
        return [0, True]


from dataclasses import dataclass, field
from typing import Optional
from detpy.models.fitness_function import FitnessFunctionBase
from detpy.models.enums.boundary_constrain import BoundaryFixing
from detpy.models.enums.optimization import OptimizationType


@dataclass
class BaseData:
    epoch: int = 100
    population_size: int = 100
    dimension: int = 10
    lb: list = field(default_factory=lambda: [-100, -100, -100, -100, -100, -100, -100, -100, -100, -100])
    ub: list = field(default_factory=lambda: [100, 100, 100, 100, 100, 100, 100, 100, 100, 100])
    mode: OptimizationType = OptimizationType.MINIMIZATION
    boundary_constraints_fun: BoundaryFixing = BoundaryFixing.RANDOM
    function: FitnessFunctionBase = None
    log_population: bool = False
    parallel_processing: Optional[list] = None


@dataclass
class DEData(BaseData):
    mutation_factor: float = 0.5
    crossover_rate: float = 0.5


@dataclass
class COMDEData(BaseData):
    mutation_factor: float = 0.1
    crossover_rate: float = 0.1


@dataclass
class DERLData(BaseData):
    mutation_factor: float = 0.1
    crossover_rate: float = 0.1


@dataclass
class NMDEData(BaseData):
    delta_f: float = 0.1
    delta_cr: float = 0.1
    sp: int = 10


@dataclass
class SADEData(BaseData):
    prob_f: float = 0.1
    prob_cr: float = 0.1


@dataclass
class EMDEData(BaseData):
    crossover_rate: float = 0.1


@dataclass
class IDEData(BaseData):
    pass


@dataclass
class DELBData(BaseData):
    crossover_rate: float = 0.1
    w_factor: float = 0.1  # control frequency of local exploration around trial and best vectors


@dataclass
class OppBasedData(BaseData):
    mutation_factor: float = 0.1
    crossover_rate: float = 0.1
    max_nfc: float = 0.1
    jumping_rate: float = 0.1


@dataclass
class DEGLData(BaseData):
    mutation_factor: float = 0.1
    crossover_rate: float = 0.1
    radius: int = 10  # neighborhood size, 2k + 1 <= NP, at least k=2
    weight: float = 0.1  # controls the balance between the exploration and exploitation


@dataclass
class JADEData(BaseData):
    archive_size: int = 10
    mutation_factor_mean: float = 0.1
    mutation_factor_std: float = 0.1
    crossover_rate_mean: float = 0.1
    crossover_rate_std: float = 0.1
    crossover_rate_low: float = 0.1
    crossover_rate_high: float = 0.1
    c: float = 0.1  # describes the rate of parameter adaptation
    p: float = 0.1  # describes the greediness of the mutation strategy


@dataclass
class AADEData(BaseData):
    mutation_factor: float = 0.1
    crossover_rate: float = 0.1


@dataclass
class EIDEData(BaseData):
    crossover_rate_min: float = 0.1
    crossover_rate_max: float = 0.1


@dataclass
class MGDEData(BaseData):
    crossover_rate: float = 0.1
    mutation_factor_f: float = 0.1
    mutation_factor_k: float = 0.1
    threshold: float = 0.1
    mu: float = 0.1


@dataclass
class FiADEData(BaseData):
    mutation_factor: float = 0.5
    crossover_rate: float = 0.5
    adaptive: bool = True


@dataclass
class ImprovedDEData(BaseData):
    mutation_factor: float = 0.1
    crossover_rate: float = 0.5



import random
import numpy as np
import copy

from detpy.models.member import Member
from detpy.models.population import Population


def em_mutation_ind(best_member: Member, better_member: Member, worst_member: Member):
    """
        Formula: v_i = x_c + F1(x_best - x_better) + F2(x_best - x_worst) + F3(x_better - x_worst)
    """
    new_member = copy.deepcopy(best_member)

    fs = np.random.uniform(size=3)
    w1, w2, w3 = get_weights()

    member_c = best_member.chromosomes * w1 + better_member.chromosomes * w2 + worst_member.chromosomes * w3
    f1_component = (best_member.chromosomes - better_member.chromosomes) * fs[0]
    f2_component = (best_member.chromosomes - worst_member.chromosomes) * fs[1]
    f3_component = (better_member.chromosomes - worst_member.chromosomes) * fs[2]
    new_member.chromosomes = member_c + f1_component + f2_component + f3_component

    return new_member


def em_mutation(population: Population):
    new_members = []
    for _ in range(population.size):
        selected_members = random.sample(population.members.tolist(), 3)
        new_member = em_mutation_ind(selected_members[0], selected_members[1], selected_members[2])
        new_members.append(new_member)

    new_population = Population(
        lb=population.lb,
        ub=population.ub,
        arg_num=population.arg_num,
        size=population.size,
        optimization=population.optimization
    )
    new_population.members = np.array(new_members)
    return new_population


def get_weights():
    p1 = 1.0
    p2 = np.random.uniform(low=0.75, high=1.0)
    p3 = np.random.uniform(low=0.5, high=p2)
    p_sum = p1 + p2 + p3

    w1 = p1 / p_sum
    w2 = p2 / p_sum
    w3 = p3 / p_sum

    return w1, w2, w3



from typing import Callable
from abc import ABC, abstractmethod


class FitnessFunctionBase(ABC):
    def __init__(self):
        self.name = ""
        self.function = None

    @abstractmethod
    def eval(self, params):
        pass


class FitnessFunction(FitnessFunctionBase):
    def __init__(self, func: Callable[..., float], custom_name=None):
        super().__init__()
        self.name = func.__name__ if custom_name is None else custom_name
        self.function = func

    def eval(self, params):
        return self.function(*params)


class FitnessFunctionOpfunu(FitnessFunctionBase):
    def __init__(self, func_type, ndim, custom_name=None):
        super().__init__()
        self.name = func_type.__name__ if custom_name is None else custom_name
        self.function = func_type(ndim=ndim)

    def eval(self, params):
        return self.function.evaluate(params)


class BenchmarkFitnessFunction(FitnessFunctionBase):
    def __init__(self, function):
        super().__init__()
        self.instance = function
        self.name = function.name

    def eval(self, params):
        return self.instance.evaluate_func(params)



from enum import Enum
import numpy as np
from detpy.models.member import Member
from detpy.models.population import Population


class BoundaryFixing(Enum):
    CLIPPING = 'clipping'
    REFLECTION = 'reflection'
    RANDOM = 'random'


def get_boundary_constraints_fun(fix_type: BoundaryFixing):
    return {
        BoundaryFixing.CLIPPING: lambda member: boundary_clipping(member),
        BoundaryFixing.REFLECTION: lambda member: boundary_reflection(member),
        BoundaryFixing.RANDOM: lambda member: boundary_random(member),
    }.get(fix_type, lambda: None)


def fix_boundary_constraints(population: Population, fix_type: BoundaryFixing):
    boundary_constraints_fun = get_boundary_constraints_fun(fix_type)

    for member in population.members:
        # Enter if member not in interval
        if not member.is_member_in_interval():
            boundary_constraints_fun(member)


# Strategies for fixing members, when they are beyond boundaries


def boundary_clipping(member: Member):
    """
    Modifies the values of `member` in-place.

    :param member: The member to be modified.
    """
    for chromosome in member.chromosomes:
        if chromosome.real_value > chromosome.ub:
            chromosome.real_value = chromosome.ub
        elif chromosome.real_value < chromosome.lb:
            chromosome.real_value = chromosome.lb


def boundary_reflection(member: Member):
    """
    Modifies the values of `member` in-place.

    :param member: The member to be modified.
    """
    for chromosome in member.chromosomes:
        if chromosome.real_value > chromosome.ub:
            chromosome.real_value = 2 * chromosome.ub - chromosome.real_value
        elif chromosome.real_value < chromosome.lb:
            chromosome.real_value = 2 * chromosome.lb - chromosome.real_value


def boundary_random(member: Member):
    """
    Modifies the values of `member` in-place.

    :param member: The member to be modified.
    """
    for chromosome in member.chromosomes:
        if chromosome.real_value < chromosome.lb or chromosome.real_value > chromosome.ub:
            chromosome.real_value = np.random.uniform(chromosome.lb, chromosome.ub)



import random
import numpy as np
import copy

from detpy.DETAlgs.methods.methods_de import mutation_ind, binomial_crossing_ind
from detpy.models.enums.optimization import OptimizationType
from detpy.models.population import Population


def nmde_mutation(population: Population, f_arr):
    new_members = []
    for i in range(population.size):
        pop_without_element = population.members.tolist()
        pop_without_element.pop(i)

        selected_members = random.sample(pop_without_element, 3)
        new_member = mutation_ind(selected_members[0], selected_members[1], selected_members[2], f_arr[i])

        new_members.append(new_member)

    new_population = Population(
        lb=population.lb,
        ub=population.ub,
        arg_num=population.arg_num,
        size=population.size,
        optimization=population.optimization
    )
    new_population.members = np.array(new_members)
    return new_population


def nmde_binomial_crossing(origin_population: Population, mutated_population: Population, cr_arr):
    if origin_population.size != mutated_population.size:
        print("Binomial_crossing: populations have different sizes")
        return None

    new_members = []
    for i in range(origin_population.size):
        new_member = binomial_crossing_ind(origin_population.members[i], mutated_population.members[i], cr_arr[i])
        new_members.append(new_member)

    new_population = Population(
        lb=origin_population.lb,
        ub=origin_population.ub,
        arg_num=origin_population.arg_num,
        size=origin_population.size,
        optimization=origin_population.optimization
    )
    new_population.members = np.array(new_members)
    return new_population


def nmde_selection(origin_population: Population, modified_population: Population):
    if origin_population.size != modified_population.size:
        print("Selection: populations have different sizes")
        return None

    if origin_population.optimization != modified_population.optimization:
        print("Selection: populations have different optimization types")
        return None

    optimization = origin_population.optimization
    new_members = []
    better_members_indexes = []
    for i in range(origin_population.size):
        if optimization == OptimizationType.MINIMIZATION:
            if origin_population.members[i] <= modified_population.members[i]:
                new_members.append(copy.deepcopy(origin_population.members[i]))
            else:
                new_members.append(copy.deepcopy(modified_population.members[i]))
                better_members_indexes.append(i)
        elif optimization == OptimizationType.MAXIMIZATION:
            if origin_population.members[i] >= modified_population.members[i]:
                new_members.append(copy.deepcopy(origin_population.members[i]))
            else:
                new_members.append(copy.deepcopy(modified_population.members[i]))
                better_members_indexes.append(i)

    new_population = Population(
        lb=origin_population.lb,
        ub=origin_population.ub,
        arg_num=origin_population.arg_num,
        size=origin_population.size,
        optimization=origin_population.optimization
    )
    new_population.members = np.array(new_members)
    return new_population, better_members_indexes


def nmde_calculate_fm_crm(set_f: set, set_cr: set):
    list_f, list_cr = list(set_f), list(set_cr)
    f_m = sum(list_f) / len(list_f) if len(list_f) > 0.0 else 0.0
    cr_m = sum(list_cr) / len(list_cr) if len(list_cr) > 0.0 else 0.0
    return f_m, cr_m


def nmde_update_f_cr(f_m, cr_m, delta_f, delta_cr):
    f_i = np.random.uniform(low=f_m - delta_f, high=f_m + delta_f)
    cr_i = np.random.uniform(low=cr_m - delta_cr, high=cr_m + delta_cr)

    if f_i > 2:
        f_i = 2
    elif f_i < 0:
        f_i = 0.2

    if cr_i > 1:
        cr_i = 1
    elif cr_i < 0:
        cr_i = 0.1

    return f_i, cr_i



import time
import typing
from dataclasses import dataclass, field

from detpy.models.member import Member
from detpy.models.population import Population


class MetricHelper:

    @staticmethod
    def calculate_start_metrics(population: Population, log_population: bool = False):
        sorted_members = population.get_best_members(population.size)
        best_inv = sorted_members[0]
        worst_inv = sorted_members[-1]

        # Metrics
        pop_mean = population.mean()
        pop_std = population.std()

        metric = Metric(
            epoch=0,
            best_individual=best_inv,
            worst_individual=worst_inv,
            population_mean=pop_mean,
            population_std=pop_std,
            execution_time=1.0
        )
        if log_population:
            metric.population = population.members

        return metric

    @staticmethod
    def calculate_metrics(population: Population, start_time, epoch, log_population: bool = False):
        sorted_members = population.get_best_members(population.size)
        best_inv = sorted_members[0]
        worst_inv = sorted_members[-1]

        # Metrics
        pop_mean = population.mean()
        pop_std = population.std()

        end_time = time.time()
        execution_time = end_time - start_time

        metric = Metric(
            epoch=epoch + 1,
            best_individual=best_inv,
            worst_individual=worst_inv,
            population_mean=pop_mean,
            population_std=pop_std,
            execution_time=execution_time
        )
        if log_population:
            metric.population = population.members

        return metric


@dataclass
class Metric:
    epoch: int
    best_individual: Member
    worst_individual: Member
    population_mean: float
    population_std: float
    execution_time: float
    population: typing.Optional[typing.List] = field(default_factory=list)



from detpy.DETAlgs.base import BaseAlg
from detpy.DETAlgs.data.alg_data import IDEData
from detpy.DETAlgs.methods.methods_de import selection, mutation
from detpy.DETAlgs.methods.methods_ide import ide_get_f, ide_get_cr, ide_binomial_crossing
from detpy.models.enums.boundary_constrain import fix_boundary_constraints


class IDE(BaseAlg):
    """
        IDE

        Links:
        https://www.scirp.org/journal/paperinformation?paperid=96749

        References:
        Ma, J. and Li, H. (2019) Research on Rosenbrock Function Optimization Problem Based on Improved Differential
        Evolution Algorithm.
        Journal of Computer and Communications, 7, 107-120. doi: 10.4236/jcc.2019.711008.
    """

    def __init__(self, params: IDEData, db_conn=None, db_auto_write=False):
        super().__init__(IDE.__name__, params, db_conn, db_auto_write)

    def next_epoch(self):
        # Calculate F and CR
        f = ide_get_f(self._epoch_number, self.num_of_epochs)
        cr_arr = ide_get_cr(self._pop)

        # New population after mutation
        v_pop = mutation(self._pop, f)

        # Apply boundary constrains on population in place
        fix_boundary_constraints(v_pop, self.boundary_constraints_fun)

        # New population after crossing
        u_pop = ide_binomial_crossing(self._pop, v_pop, cr_arr)

        # Update values before selection
        u_pop.update_fitness_values(self._function.eval, self.parallel_processing)

        # Select new population
        new_pop = selection(self._pop, u_pop)

        # Override data
        self._pop = new_pop

        self._epoch_number += 1



import copy
import random
import numpy as np

from detpy.DETAlgs.methods.methods_de import mutation_ind, binomial_crossing_ind
from detpy.models.enums.optimization import OptimizationType
from detpy.models.population import Population


def sade_mutation(population: Population, f_arr):
    new_members = []
    for i in range(population.size):
        selected_members = random.sample(population.members.tolist(), 3)
        new_member = mutation_ind(selected_members[0], selected_members[1], selected_members[2], f_arr[i])
        new_members.append(new_member)

    new_population = Population(
        lb=population.lb,
        ub=population.ub,
        arg_num=population.arg_num,
        size=population.size,
        optimization=population.optimization
    )
    new_population.members = np.array(new_members)
    return new_population


def sade_binomial_crossing(origin_population: Population, mutated_population: Population, cr_arr):
    if origin_population.size != mutated_population.size:
        print("Binomial_crossing: populations have different sizes")
        return None

    new_members = []
    for i in range(origin_population.size):
        new_member = binomial_crossing_ind(origin_population.members[i], mutated_population.members[i], cr_arr[i])
        new_members.append(new_member)

    new_population = Population(
        lb=origin_population.lb,
        ub=origin_population.ub,
        arg_num=origin_population.arg_num,
        size=origin_population.size,
        optimization=origin_population.optimization
    )
    new_population.members = np.array(new_members)
    return new_population


def sade_selection(origin_population: Population, modified_population: Population, f_arr, cr_arr, prob_f, prob_cr):
    if origin_population.size != modified_population.size:
        print("Selection: populations have different sizes")
        return None

    if origin_population.optimization != modified_population.optimization:
        print("Selection: populations have different optimization types")
        return None

    optimization = origin_population.optimization
    new_members = []
    for i in range(origin_population.size):
        if optimization == OptimizationType.MINIMIZATION:
            if origin_population.members[i] <= modified_population.members[i]:
                new_members.append(copy.deepcopy(origin_population.members[i]))
                if np.random.uniform() < prob_f:
                    f_arr[i] = np.random.uniform()
            else:
                new_members.append(copy.deepcopy(modified_population.members[i]))
                if np.random.uniform() >= prob_cr:
                    cr_arr[i] = np.random.uniform()
        elif optimization == OptimizationType.MAXIMIZATION:
            if origin_population.members[i] >= modified_population.members[i]:
                new_members.append(copy.deepcopy(origin_population.members[i]))
            else:
                new_members.append(copy.deepcopy(modified_population.members[i]))
                f_arr[i] = np.random.uniform() if np.random.uniform() < prob_f else f_arr[i]
                cr_arr[i] = np.random.uniform() if np.random.uniform() < prob_cr else cr_arr[i]

    new_population = Population(
        lb=origin_population.lb,
        ub=origin_population.ub,
        arg_num=origin_population.arg_num,
        size=origin_population.size,
        optimization=origin_population.optimization
    )
    new_population.members = np.array(new_members)
    return new_population, f_arr, cr_arr



import random
from detpy.DETAlgs.base import BaseAlg
from detpy.DETAlgs.data.alg_data import EIDEData
from detpy.DETAlgs.methods.methods_de import mutation, binomial_crossing, selection
from detpy.DETAlgs.methods.methods_eide import eide_adopt_parameters
from detpy.models.enums.boundary_constrain import fix_boundary_constraints


class EIDE(BaseAlg):
    """
        EIDE

        Links:
        https://ieeexplore.ieee.org/document/6390324

        References:
        Z. Dexuan and G. Liqun, "An efficient improved differential evolution algorithm,"
        Proceedings of the 31st Chinese Control Conference, Hefei, China, 2012, pp. 2385-2390.
    """
    def __init__(self, params: EIDEData, db_conn=None, db_auto_write=False):
        super().__init__(EIDE.__name__, params, db_conn, db_auto_write)

        self.mutation_factor = random.uniform(0, 0.6)
        self.crossover_rate = params.crossover_rate_min
        self.crossover_rate_min = params.crossover_rate_min
        self.crossover_rate_max = params.crossover_rate_max
        self.generation = None

    def next_epoch(self):
        # New population after mutation
        v_pop = mutation(self._pop, f=self.mutation_factor)

        # Apply boundary constrains on population in place
        fix_boundary_constraints(v_pop, self.boundary_constraints_fun)

        # New population after crossing
        u_pop = binomial_crossing(self._pop, v_pop, cr=self.crossover_rate)

        # Update values before selection
        u_pop.update_fitness_values(self._function.eval, self.parallel_processing)

        # Select new population
        new_pop = selection(self._pop, u_pop)

        self.mutation_factor, self.crossover_rate = eide_adopt_parameters(self.crossover_rate_min,
                                                                          self.crossover_rate_max, self._epoch_number,
                                                                          self.num_of_epochs)

        # Override data
        self._pop = new_pop

        self._epoch_number += 1



import copy
import time
import traceback
from abc import ABC, abstractmethod
from statistics import mean, stdev
from tqdm import tqdm
import numpy as np
import detpy
from detpy.database.database_connector import SQLiteConnector
from detpy.DETAlgs.data.alg_data import BaseData
from detpy.helpers.database_helper import get_table_name, format_individuals
from detpy.helpers.metric_helper import MetricHelper
from detpy.models.algorithm_result import AlgorithmResult
from detpy.models.fitness_function import FitnessFunctionBase
from detpy.models.population import Population
from detpy.helpers.logger import Logger


def example_function(x1, x2, x3, x4, x5, x6, x7, x8, x9, x10):
    return (x1 - 1) ** 2 + (x2 - 2) ** 2 + (x3 - 3) ** 2 + (x4 - 4) ** 2 + (x5 - 5) ** 2 + \
           (x6 - 6) ** 2 + (x7 - 7) ** 2 + (x8 - 8) ** 2 + (x9 - 9) ** 2 + (x10 - 10) ** 2


class BaseAlg(ABC):
    def __init__(self, name, params: BaseData, db_conn=None, db_auto_write=False, verbose=True):
        self.name = name
        self._epoch_number = 0

        self._origin_pop = None
        self._pop = None

        self.num_of_epochs = params.epoch
        self.population_size = params.population_size
        self.nr_of_args = params.dimension
        self.lb = params.lb
        self.ub = params.ub
        self.mode = params.mode
        self.boundary_constraints_fun = params.boundary_constraints_fun

        if params.function is None:
            self._function = detpy.FitnessFunction(func=example_function)
        else:
            self._function: FitnessFunctionBase = params.function

        self._database = SQLiteConnector(db_conn) if db_conn is not None else None
        self.db_auto_write = db_auto_write
        self.log_population = params.log_population
        self.parallel_processing = params.parallel_processing
        self.database_table_name = None
        self.db_writing_interval = 50

        # Use Logger for output control
        self.logger = Logger(verbose)
        self._initialize()

    @abstractmethod
    def next_epoch(self):
        pass

    def _initialize(self):
        population = Population(
            lb=self.lb,
            ub=self.ub,
            arg_num=self.nr_of_args,
            size=self.population_size,
            optimization=self.mode
        )
        population.generate_population()
        population.update_fitness_values(self._function.eval, self.parallel_processing)

        self._origin_pop = population
        self._pop = copy.deepcopy(population)

        # Creating table
        self._database.connect()
        table_name = get_table_name(
            func_name="aa",
            alg_name=self.name,
            nr_of_args=self.nr_of_args,
            pop_size=self.population_size
        )
        self.database_table_name = self._database.create_table(table_name)
        self._database.close()

    def run(self):
        epoch_metrics = []
        best_fitness_values = []
        avg_fitness_values = []
        std_fitness_values = []

        # Calculate metrics
        epoch_metric = MetricHelper.calculate_start_metrics(self._pop, self.log_population)
        epoch_metrics.append(epoch_metric)

        start_time = time.time()
        end_index = 0
        bar_format = '{l_bar}{bar}{r_bar}\n'

        for epoch in tqdm(range(self.num_of_epochs), desc=f"{self.name}", unit="epoch", bar_format=bar_format):
            best_member = self._pop.get_best_members(1)[0]
            best_fitness_values.append(best_member.fitness_value)

            try:
                self.next_epoch()

                # Calculate metrics
                epoch_metric = MetricHelper.calculate_metrics(self._pop, start_time, epoch, self.log_population)
                epoch_metrics.append(epoch_metric)

                avg_fitness = mean(member.fitness_value for member in self._pop.members)
                avg_fitness_values.append(avg_fitness)

                std_fitness = stdev(member.fitness_value for member in self._pop.members)
                std_fitness_values.append(std_fitness)

                self.logger.log(f"Epoch {epoch + 1}/{self.num_of_epochs}, Best Fitness: {best_member.fitness_value}, "
                                f"Best Individual: {[member.real_value for member in best_member.chromosomes]}, "
                                f"Avg: {avg_fitness}, Std: {std_fitness}")

                # Saving after each 50 epochs
                if epoch > 0 and epoch % self.db_writing_interval == 0:
                    end_index = epoch + 1
                    start_index = 0 if epoch == self.db_writing_interval else end_index - self.db_writing_interval
                    if self._database is not None and self.db_auto_write:
                        try:
                            self.write_results_to_database(epoch_metrics[start_index:end_index])
                        except:
                            self.logger.log('An unexpected error occurred while writing to the database.')
            except Exception as e:
                traceback.print_exc()
                self.logger.log(f'An unexpected error occurred during calculation: {e}')
                return epoch_metrics

        end_time = time.time()
        execution_time = end_time - start_time
        self.logger.log(f'Function: {self._function.name}, Dimension: {self.nr_of_args},'
                        f' Execution time: {round(execution_time, 2)} seconds')

        avg_fitness = np.mean(best_fitness_values)
        std_fitness = np.std(best_fitness_values)
        best_solution = self._pop.get_best_members(1)[0]

        self.logger.log(f"Average Best Fitness: {avg_fitness}, Standard Deviation of Fitness: {std_fitness}")
        self.logger.log(f"Best Solution: {best_solution}")

        # Writing to database
        if self._database is not None and not self.db_auto_write:
            try:
                self.write_results_to_database(epoch_metrics)
            except Exception as e:
                self.logger.log(f'An unexpected error occurred while writing to the database: {e}')
        elif self._database is not None and self.db_auto_write:
            try:
                self.write_results_to_database(epoch_metrics[end_index:])
            except Exception as e:
                self.logger.log(f'An unexpected error occurred while writing to the database: {e}')

        result = AlgorithmResult(
            epoch_metrics=epoch_metrics,
            avg_fitness=avg_fitness,
            std_fitness=std_fitness,
            best_solution=best_solution
        )

        result.plot_results(best_fitness_values, avg_fitness_values, std_fitness_values, self.num_of_epochs,
                            method_name=self.name)

        return result

    def write_results_to_database(self, results_data):
        self.logger.log(f'Writing to Database...')

        # Check if database is present
        if self._database is None or self.database_table_name is None:
            self.logger.log(f"There is no database.")
            return

        # Connect to database
        self._database.connect()

        # Inserting data into database
        formatted_best_individuals = format_individuals(results_data)
        self._database.insert_multiple_best_individuals(self.database_table_name, formatted_best_individuals)

        self._database.close()



import numpy as np
from math import floor
import random
import copy

from detpy.models.population import Population
from detpy.models.member import Member
from detpy.models.enums.mutation import mutation_curr_to_best_1
from detpy.models.enums.optimization import OptimizationType
from detpy.DETAlgs.methods.methods_de import binomial_crossing_ind


def jade_mutation(population: Population, mutation_factors: np.ndarray[float], p_best: float,
                  archive: list[Member] = []) -> Population:

    pop_members_list = population.members.tolist()
    possible_members_list = pop_members_list + archive if archive else pop_members_list

    sort_reversed = False if population.optimization == OptimizationType.MINIMIZATION else True
    sorted_members = sorted(possible_members_list, key=lambda x: x.fitness_value, reverse=sort_reversed)
    p_best_members = sorted_members[:floor(p_best * population.size)]

    new_members = []
    for (i, base_member) in enumerate(pop_members_list):
        best_member = random.choice(p_best_members)

        indices = list(range(population.size))
        indices.remove(i)
        idx_r1 = random.choice(indices)
        x_r1 = pop_members_list[idx_r1]

        indices = list(range(len(possible_members_list)))
        indices.remove(i)
        indices.remove(idx_r1)
        idx_r2 = random.choice(indices)
        x_r2 = possible_members_list[idx_r2]

        new_member = mutation_curr_to_best_1(base_member, best_member, [x_r1, x_r2], mutation_factors[i])
        new_members.append(new_member)

    new_population = Population(
        lb=population.lb,
        ub=population.ub,
        arg_num=population.arg_num,
        size=population.size,
        optimization=population.optimization
    )
    new_population.members = np.array(new_members)

    return new_population


def jade_binomial_crossing(origin_population: Population, mutated_population: Population,
                           crossover_rates: np.ndarray[float]) -> Population | None:
    if origin_population.size != mutated_population.size:
        print("Binomial_crossing: populations have different sizes")
        return None

    new_members = []
    for i in range(origin_population.size):
        new_member = binomial_crossing_ind(origin_population.members[i], mutated_population.members[i], crossover_rates[i])
        new_members.append(new_member)

    new_population = Population(
        lb=origin_population.lb,
        ub=origin_population.ub,
        arg_num=origin_population.arg_num,
        size=origin_population.size,
        optimization=origin_population.optimization
    )
    new_population.members = np.array(new_members)
    return new_population


def jade_selection(origin_population: Population, modified_population: Population,
                   mutation_factors: np.ndarray[float], crossover_rates: np.ndarray[float],
                   success_mutation_factors: list[float], success_crossover_rates: list[float],
                   archive: list[Member] = []) -> Population | None:
    if origin_population.size != modified_population.size != len(mutation_factors) != len(crossover_rates):
        print("Selection: parameters lists have different sizes")
        return None

    if origin_population.optimization != modified_population.optimization:
        print("Selection: populations have different optimization types")
        return None

    optimization = origin_population.optimization
    new_members = []
    for i in range(origin_population.size):
        if optimization == OptimizationType.MINIMIZATION:
            if origin_population.members[i] <= modified_population.members[i]:
                new_members.append(copy.deepcopy(origin_population.members[i]))
            else:
                new_members.append(copy.deepcopy(modified_population.members[i]))
                archive.append(copy.deepcopy(origin_population.members[i]))
                success_mutation_factors.append(mutation_factors[i])
                success_crossover_rates.append(crossover_rates[i])

        elif optimization == OptimizationType.MAXIMIZATION:
            if origin_population.members[i] >= modified_population.members[i]:
                new_members.append(copy.deepcopy(origin_population.members[i]))
            else:
                new_members.append(copy.deepcopy(modified_population.members[i]))
                archive.append(copy.deepcopy(origin_population.members[i]))
                success_mutation_factors.append(mutation_factors[i])
                success_crossover_rates.append(crossover_rates[i])

    new_population = Population(
        lb=origin_population.lb,
        ub=origin_population.ub,
        arg_num=origin_population.arg_num,
        size=origin_population.size,
        optimization=origin_population.optimization
    )
    new_population.members = np.array(new_members)

    return new_population


def jade_reduce_archive(archive: list[Member], archive_size: int):
    if archive_size == 0:
        archive.clear()

    reduce_num = len(archive) - archive_size
    for _ in range(reduce_num):
        idx = random.randrange(len(archive))
        archive.pop(idx)


def jade_adapt_crossover_rates(c: float, cr_mean: float, cr_std: float, cr_low: float, cr_high: float, size: int,
                               success_crossover_rates: list[float]) -> tuple[np.ndarray[float], float]:
    if success_crossover_rates:
        a_mean = sum(success_crossover_rates) / len(success_crossover_rates)
        cr_mean = (1 - c) * cr_mean + c * a_mean

    new_crossover_rates = draw_norm_dist_within_bounds(cr_mean, cr_std, cr_low, cr_high, size)
    success_crossover_rates.clear()
    return new_crossover_rates, cr_mean


def jade_adapt_mutation_factors(c: float, f_mean: float, f_std: float, size: int,
                                success_mutation_factors: list[float]) -> tuple[np.ndarray[float], float]:
    if success_mutation_factors:
        f_mean = (1 - c) * f_mean + c * lehmer_mean(success_mutation_factors)

    new_mutation_factors = draw_cauchy_dist_within_bounds(f_mean, f_std, size)
    success_mutation_factors.clear()
    return new_mutation_factors, f_mean


def draw_norm_dist_within_bounds(mean: float, std: float, low: float, high: float, arr_size: int) -> np.ndarray[float]:
    """
    Draw numbers from normal distributions within bounds (low, high].
    """
    values = np.clip(np.random.normal(loc=mean, scale=std, size=arr_size), low, high)

    while len(values) < arr_size:
        new_values = np.clip(np.random.normal(loc=mean, scale=std, size=arr_size - len(values)), low, high)
        values = np.concatenate((values, new_values))

    return values[:arr_size]


def draw_cauchy_dist_within_bounds(mean: float, std: float, arr_size: int) -> np.ndarray[float]:
    values = []

    while len(values) < arr_size:
        value = mean + std * np.random.standard_cauchy()

        if value >= 1:
            values.append(1.0)
        elif 0 < value < 1:
            values.append(value)

    return np.array(values)


def lehmer_mean(mut_factors: list[float]) -> float:
    numbers_array = np.array(mut_factors)

    numerator = np.sum(numbers_array ** 2)
    denominator = np.sum(numbers_array)

    return numerator / denominator



from detpy.DETAlgs.base import BaseAlg
from detpy.DETAlgs.data.alg_data import COMDEData
from detpy.DETAlgs.methods.methods_comde import calculate_cr, comde_mutation
from detpy.DETAlgs.methods.methods_de import binomial_crossing, selection
from detpy.models.enums.boundary_constrain import fix_boundary_constraints


class COMDE(BaseAlg):
    """
        COMDE

        Links:
        https://www.sciencedirect.com/science/article/pii/S0020025512000278

        References:
        Mohamed, A. W., & Sabry, H. Z. (2012). Constrained optimization based on modified differential evolution algorithm.
        In Information Sciences (Vol. 194, pp. 171–208). Elsevier BV. https://doi.org/10.1016/j.ins.2012.01.008

    """

    def __init__(self, params: COMDEData, db_conn=None, db_auto_write=False):
        super().__init__(COMDE.__name__, params, db_conn, db_auto_write)

        self.mutation_factor = params.mutation_factor  # F
        self.crossover_rate = params.crossover_rate  # Cr

    def next_epoch(self):
        # Calculate not constant cr depend on generation number
        cr = calculate_cr(self._epoch_number, self.num_of_epochs)

        # New population after mutation
        v_pop = comde_mutation(self._pop)

        # Apply boundary constrains on population in place
        fix_boundary_constraints(v_pop, self.boundary_constraints_fun)

        # New population after crossing
        u_pop = binomial_crossing(self._pop, v_pop, cr=cr)

        # Update values before selection
        u_pop.update_fitness_values(self._function.eval, self.parallel_processing)

        # Select new population
        new_pop = selection(self._pop, u_pop)

        # Override data
        self._pop = new_pop

        self._epoch_number += 1



from detpy.DETAlgs.base import BaseAlg
from detpy.DETAlgs.methods.methods_fiade import mutation, crossing, selection, adapt_parameters
from detpy.models.enums.boundary_constrain import fix_boundary_constraints
from detpy.DETAlgs.data.alg_data import FiADEData


class FiADE(BaseAlg):
    """
        FiADE

        Links:
        https://www.sciencedirect.com/science/article/abs/pii/S0020025511001381

        References:
        Arnob Ghosh , Swagatam Das , Aritra Chowdhury , Ritwik Giri (2011)
        An Improved Differential Evolution Algorithm with Fitness-Based Adaptation of the Control Parameters.
        Volume 181, Issue 18, 15 September 2011, Pages 3749-3765
        doi: 10.1016/j.ins.2011.03.010
    """

    def __init__(self, params: FiADEData, db_conn=None, db_auto_write=False):
        super().__init__(FiADE.__name__, params, db_conn, db_auto_write)

        # Initialize mutation factors and crossover rates for the population
        self.mutation_factors = [[params.mutation_factor, False] for _ in range(params.population_size)]
        self.crossover_rates = [[params.crossover_rate, False] for _ in range(params.population_size)]

    def next_epoch(self):
        # Mutation
        v_pop = mutation(self._pop, self.mutation_factors)

        # Apply boundary constraints
        fix_boundary_constraints(v_pop, self.boundary_constraints_fun)

        # Crossing
        u_pop = crossing(self._pop, v_pop, self.crossover_rates)

        # Update values before selection
        u_pop.update_fitness_values(self._function.eval, self.parallel_processing)

        # Selection
        new_pop = selection(self._pop, u_pop, self.mutation_factors, self.crossover_rates)

        # Adapt parameters based on fitness
        adapt_parameters(new_pop, self.mutation_factors, self.crossover_rates)

        # Override population with the newly selected population
        self._pop = new_pop
        self._epoch_number += 1



import sqlite3


class BaseSQLiteConnector:
    def __init__(self, database):
        self.database = database
        self.connection = None
        self.cursor = None

    def connect(self):
        self.connection = sqlite3.connect(self.database)
        self.cursor = self.connection.cursor()

    def execute_query(self, query, params=None):
        if params:
            self.cursor.execute(query, params)
        else:
            self.cursor.execute(query)

    def executemany_query(self, query, params=None):
        if params:
            self.cursor.executemany(query, params)
        else:
            self.cursor.executemany(query)

    def fetch_all(self):
        return self.cursor.fetchall()

    def fetch_one(self):
        return self.cursor.fetchone()

    def commit(self):
        self.connection.commit()

    def close(self):
        self.cursor.close()
        self.connection.close()


class SQLiteConnector(BaseSQLiteConnector):
    def __init__(self, database):
        super().__init__(database)

    def create_table(self, table_name):
        new_table_name = self.find_valid_table_name(table_name)

        create_table_query = f'''
            CREATE TABLE {new_table_name} (
                id INTEGER PRIMARY KEY,
                epoch INTEGER,
                argumentsBest TEXT,
                fitnessValueBest REAL,
                argumentsWorst TEXT,
                fitnessValueWorst REAL,
                mean REAL,
                std REAL,
                calculationTime REAL,
                population TEXT
            )
        '''

        self.execute_query(create_table_query)
        self.commit()

        return new_table_name

    def find_valid_table_name(self, table_name):
        query = "SELECT name FROM sqlite_master WHERE type='table' AND name LIKE ?"
        self.execute_query(query, (table_name + '%',))
        results = self.fetch_all()
        tables = [row[0] for row in results]

        return f"{table_name}_{len(tables) + 1}"

    def find_existing_table_name(self, table_name):
        query = "SELECT name FROM sqlite_master WHERE type='table' AND name LIKE ?"
        self.execute_query(query, (table_name + '%',))
        results = self.fetch_all()
        tables = [row[0] for row in results]

        return f"{table_name}_{len(tables)}"

    def insert_multiple_best_individuals(self, table_name, params):
        insert_query = f'''
        INSERT INTO {table_name} (epoch, argumentsBest, fitnessValueBest, argumentsWorst, fitnessValueWorst, mean, std, calculationTime, population)
                       VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)
        '''

        self.executemany_query(insert_query, params)
        self.commit()

    def get_results(self, table_name):
        query = f'''
        SELECT * FROM {table_name}
        '''

        self.execute_query(query)
        return self.fetch_all()



from detpy.DETAlgs.base import BaseAlg
from detpy.DETAlgs.data.alg_data import EMDEData
from detpy.DETAlgs.methods.methods_de import binomial_crossing, selection
from detpy.DETAlgs.methods.methods_emde import em_mutation
from detpy.models.enums.boundary_constrain import fix_boundary_constraints


class EMDE(BaseAlg):
    """
        EMDE

        Links:
        https://link.springer.com/article/10.1007/s13042-015-0479-6

        References:
        Mohamed, A.W. An efficient modified differential evolution algorithm for solving constrained non-linear
        integer and mixed-integer global optimization problems.
        Int. J. Mach. Learn. & Cyber. 8, 989–1007 (2017). https://doi.org/10.1007/s13042-015-0479-6
    """

    def __init__(self, params: EMDEData, db_conn=None, db_auto_write=False):
        super().__init__(EMDE.__name__, params, db_conn, db_auto_write)

        self.crossover_rate = params.crossover_rate  # Cr

    def next_epoch(self):
        # Calculate not constant cr depend on generation number
        v_pop = em_mutation(self._pop)

        # Apply boundary constrains on population in place
        fix_boundary_constraints(v_pop, self.boundary_constraints_fun)

        # New population after crossing
        u_pop = binomial_crossing(self._pop, v_pop, self.crossover_rate)

        # Update values before selection
        u_pop.update_fitness_values(self._function.eval, self.parallel_processing)

        # Select new population
        new_pop = selection(self._pop, u_pop)

        # Override data
        self._pop = new_pop

        self._epoch_number += 1



from detpy.DETAlgs.base import BaseAlg
from detpy.DETAlgs.data.alg_data import OppBasedData
from detpy.DETAlgs.methods.methods_opposition_based import opp_based_generation_jumping
from detpy.DETAlgs.methods.methods_de import mutation, binomial_crossing, selection
from detpy.models.enums.boundary_constrain import fix_boundary_constraints


class OppBasedDE(BaseAlg):
    """
        OppBasedDE

        Links:
        https://ieeexplore.ieee.org/document/4358759

        References:
        S. Rahnamayan, H. R. Tizhoosh and M. M. A. Salama, "Opposition-Based Differential Evolution,"
        in IEEE Transactions on Evolutionary Computation, vol. 12, no. 1, pp. 64-79, Feb. 2008,
        doi: 10.1109/TEVC.2007.894200.
    """

    def __init__(self, params: OppBasedData, db_conn=None, db_auto_write=False):
        super().__init__(OppBasedDE.__name__, params, db_conn, db_auto_write)

        self.mutation_factor = params.mutation_factor  # F
        self.crossover_rate = params.crossover_rate  # Cr
        self.nfc = 0  # number of function calls
        self.max_nfc = params.max_nfc
        self.jumping_rate = params.jumping_rate

    def next_epoch(self):
        # New population after mutation
        v_pop = mutation(self._pop, f=self.mutation_factor)

        # Apply boundary constrains on population in place
        fix_boundary_constraints(v_pop, self.boundary_constraints_fun)

        # New population after crossing
        u_pop = binomial_crossing(self._pop, v_pop, cr=self.crossover_rate)

        # Update values before selection
        u_pop.update_fitness_values(self._function.eval)
        self.nfc += self.population_size

        # Select new population
        new_pop = selection(self._pop, u_pop)

        # Generation jumping
        if opp_based_generation_jumping(new_pop, self.jumping_rate, self._function.eval):
            self.nfc += self.population_size

        # Override data
        self._pop = new_pop

        self._epoch_number += 1



from detpy.DETAlgs.base import BaseAlg
from detpy.DETAlgs.data.alg_data import DEGLData
from detpy.DETAlgs.methods.methods_de import binomial_crossing, selection
from detpy.DETAlgs.methods.methods_degl import degl_mutation, degl_adapt_weight
from detpy.models.enums.boundary_constrain import fix_boundary_constraints


class DEGL(BaseAlg):
    """
        DEGL

        Links:
        https://ieeexplore.ieee.org/document/5089881/

        References:
        S. Das, A. Abraham, U. K. Chakraborty and A. Konar,
        "Differential Evolution Using a Neighborhood-Based Mutation Operator,"
        in IEEE Transactions on Evolutionary Computation, vol. 13, no. 3, pp. 526-553, June 2009,
        doi: 10.1109/TEVC.2008.2009457.
    """

    def __init__(self, params: DEGLData, db_conn=None, db_auto_write=False):
        super().__init__(DEGL.__name__, params, db_conn, db_auto_write)

        self.mutation_factor = params.mutation_factor  # F
        self.crossover_rate = params.crossover_rate  # Cr
        self.radius = params.radius
        self.weight = 0

    def next_epoch(self):
        # New population after mutation
        v_pop = degl_mutation(self._pop, self.radius, self.mutation_factor, self.weight)

        # Apply boundary constrains on population in place
        fix_boundary_constraints(v_pop, self.boundary_constraints_fun)

        # New population after crossing
        u_pop = binomial_crossing(self._pop, v_pop, cr=self.crossover_rate)

        # Update values before selection
        u_pop.update_fitness_values(self._function.eval, self.parallel_processing)

        # Select new population
        new_pop = selection(self._pop, u_pop)

        self.weight = degl_adapt_weight(self._epoch_number, self.num_of_epochs)

        # Override data
        self._pop = new_pop

        self._epoch_number += 1



from detpy.DETAlgs.base import BaseAlg
from detpy.DETAlgs.data.alg_data import DEData
from detpy.DETAlgs.methods.methods_de import mutation, binomial_crossing, selection
from detpy.models.enums.boundary_constrain import fix_boundary_constraints


class DE(BaseAlg):
    """
        The original version of different evolution

        Links:
        https://link.springer.com/article/10.1023/A:1008202821328

        References:
        Storn, R., Price, K.
        Differential Evolution – A Simple and Efficient Heuristic for global Optimization over Continuous Spaces.
        Journal of Global Optimization 11, 341–359 (1997).
        https://doi.org/10.1023/A:1008202821328

    """

    def __init__(self, params: DEData, db_conn=None, db_auto_write=False):
        super().__init__(DE.__name__, params, db_conn, db_auto_write)

        self.mutation_factor = params.mutation_factor  # F
        self.crossover_rate = params.crossover_rate  # Cr

    def next_epoch(self):
        # New population after mutation
        v_pop = mutation(self._pop, f=self.mutation_factor)

        # Apply boundary constrains on population in place
        fix_boundary_constraints(v_pop, self.boundary_constraints_fun)

        # New population after crossing
        u_pop = binomial_crossing(self._pop, v_pop, cr=self.crossover_rate)

        # Update values before selection
        u_pop.update_fitness_values(self._function.eval, self.parallel_processing)

        # Select new population
        new_pop = selection(self._pop, u_pop)

        # Override data
        self._pop = new_pop

        self._epoch_number += 1



import copy
import random
import numpy as np

from detpy.models.enums.mutation import mutation_rand_1
from detpy.models.member import Member
from detpy.models.population import Population


def mgde_mutation(population: Population, curr_generation: int, max_generation: float, mutation_factor_f: float,
                  mutation_factor_k: float) -> Population:
    pop_members_list = population.members.tolist()
    best_member = population.get_best_members(1)[0]

    new_members = []
    for i, member in enumerate(population.members):
        indices = list(range(population.size))
        indices.remove(i)
        selected_idxs = random.sample(indices, 3)

        donors = [pop_members_list[idx] for idx in selected_idxs]
        if random.random() > (curr_generation - 1) / (max_generation - 1):
            new_members.append(mutation_rand_1(donors, mutation_factor_f))
        else:
            new_members.append(mutation_curr_to_best_1_modified(member, best_member, donors,
                                                                mutation_factor_f, mutation_factor_k))

    new_population = Population(
        lb=population.lb,
        ub=population.ub,
        arg_num=population.arg_num,
        size=population.size,
        optimization=population.optimization
    )
    new_population.members = np.array(new_members)
    return new_population


def mgde_adapt_threshold(population: Population, threshold: float, mu: float, func):
    sorted_members = population.get_best_members(population.size)
    best_member = sorted_members[0]

    if best_member.fitness_value < threshold:
        threshold /= 10
        if random.random() > 0.5:
            worst_members = sorted_members[-5:]
            worst_member = worst_members[random.randint(0, 4)]
            member_chromosome_values = np.array([chromosome.real_value for chromosome in worst_member.chromosomes])
            new_chromosome_values = mu * member_chromosome_values * (1 - member_chromosome_values)
            for i in range(worst_member.args_num):
                worst_member.chromosomes[i].real_value = new_chromosome_values[i]
        else:
            member = sorted_members[random.randint(1, population.size - 1)]
            d = random.randint(0, member.args_num - 1)
            n = random.randint(0, 1)
            h = np.random.normal(loc=0.5, scale=0.3)

            temp = copy.deepcopy(member)
            temp.chromosomes[d] = member.chromosomes[d] + (
                        member.chromosomes[d] - best_member.chromosomes[d]) * h * (-1) ** n

            temp.calculate_fitness_fun(func)
            if temp < best_member:
                sorted_members[0] = temp


def mutation_curr_to_best_1_modified(base_member: Member, best_member: Member, members: list[Member], f: float,
                                     k: float) -> Member:
    """
        Formula: v_ij = x_r1 + F(x_r2 - x_r3) + K(x_best - x_base)
    """
    new_member = copy.deepcopy(base_member)
    new_member.chromosomes = members[0].chromosomes + (members[1].chromosomes - members[2].chromosomes) * f + (
            best_member.chromosomes - base_member.chromosomes) * k
    return new_member



import random
import numpy as np

from detpy.models.population import Population
from detpy.models.member import Member
from detpy.models.enums.mutation import mutation_curr_to_best_1


def generate_local_donor(base_member: Member, neighborhood: Population, f: float) -> Member:
    best_local = neighborhood.get_best_members(1)[0]
    selected_members = random.sample(neighborhood.members.tolist(), 2)
    local_donor = mutation_curr_to_best_1(base_member, best_local, selected_members, f)
    return local_donor


def combine_donors(global_donor: Member, local_donor: Member, weight: float) -> Member:
    return global_donor * weight + local_donor * (1 - weight)


def degl_mutation(population: Population, radius: int, f: float, weight: float) -> Population:
    pop_members_list = population.members.tolist()
    best_global = population.get_best_members(1)[0]

    new_members = []
    population_indices = list(range(population.size))
    topology_indices = population_indices[-radius:] + population_indices + population_indices[:radius]
    for i, member in enumerate(population.members):
        # Generate local donor
        local_indices = topology_indices[i:i + 2 * radius + 1]
        local_indices.remove(i)
        neighborhood_members = [pop_members_list[i] for i in local_indices]
        neighborhood = Population(
            lb=population.lb,
            ub=population.ub,
            arg_num=population.arg_num,
            size=2 * radius,
            optimization=population.optimization
        )
        neighborhood.members = np.array(neighborhood_members)
        local_donor = generate_local_donor(member, neighborhood, f)

        # Generating global donor
        global_indices = list(range(population.size))
        global_indices.remove(i)
        selected_indices = random.sample(global_indices, 2)
        global_members = [pop_members_list[i] for i in selected_indices]
        global_donor = mutation_curr_to_best_1(member, best_global, global_members, f)

        # Combining donors into new member
        new_member = combine_donors(global_donor, local_donor, weight)
        new_members.append(new_member)

    new_population = Population(
        lb=population.lb,
        ub=population.ub,
        arg_num=population.arg_num,
        size=population.size,
        optimization=population.optimization
    )
    new_population.members = np.array(new_members)

    return new_population


def degl_adapt_weight(current_generation: int, max_generation: int) -> float:
    return current_generation / max_generation



import random
import numpy as np

from detpy.DETAlgs.methods.methods_de import mutation_ind
from detpy.models.population import Population


def comde_mutation(population: Population):
    new_members = []
    for _ in range(population.size):
        f_l = np.random.uniform()                   # random in (0, 1]
        f_g = np.random.uniform(low=-1, high=1)     # random in (−1, 0) ∪ (0, 1)
        while f_g == 0 or f_l == 0:  # not pleasant
            f_l = np.random.uniform()
            f_g = np.random.uniform(low=-1, high=1)

        if np.random.uniform() <= 0.5:
            # Select worst and best members
            sorted_members = population.get_best_members(population.size)
            best_member, worst_member = sorted_members[0], sorted_members[-1]
            selected_member = random.sample(sorted_members[1:-1].tolist(), 1)[0]
            new_member = mutation_ind(selected_member, best_member, worst_member, f_l)
        else:
            # Select members by random
            selected_members = random.sample(population.members.tolist(), 3)
            new_member = mutation_ind(selected_members[0], selected_members[1], selected_members[2], f_g)

        new_members.append(new_member)

    new_population = Population(
        lb=population.lb,
        ub=population.ub,
        arg_num=population.arg_num,
        size=population.size,
        optimization=population.optimization
    )
    new_population.members = np.array(new_members)
    return new_population


def calculate_cr(curr_gen, max_gen, cr_min=0.5, cr_max=0.95, k=4):
    cr = cr_max + (cr_min - cr_max) * (1 - curr_gen / max_gen) ** k
    return cr



import random


def eide_adopt_parameters(crossover_rate_min: float, crossover_rate_max: float, curr_generation: int,
                          max_generation: int) -> tuple[float, float]:
    new_f = random.uniform(0, 0.6)
    new_cr = crossover_rate_min + (crossover_rate_max-crossover_rate_min) * curr_generation / max_generation

    return new_f, new_cr



from detpy.DETAlgs.base import BaseAlg
from detpy.DETAlgs.data.alg_data import JADEData
from detpy.DETAlgs.methods.methods_jade import jade_adapt_mutation_factors, jade_binomial_crossing, \
    jade_adapt_crossover_rates, jade_mutation, jade_selection, jade_reduce_archive, draw_norm_dist_within_bounds, \
    draw_cauchy_dist_within_bounds
from detpy.models.enums.boundary_constrain import fix_boundary_constraints


class JADE(BaseAlg):
    """
        JADE

        Links:
        https://ieeexplore.ieee.org/document/5208221

        References:
        J. Zhang and A. C. Sanderson, "JADE: Adaptive Differential Evolution With Optional External Archive,"
        in IEEE Transactions on Evolutionary Computation, vol. 13, no. 5, pp. 945-958, Oct. 2009,
        doi: 10.1109/TEVC.2009.2014613.
    """

    def __init__(self, params: JADEData, db_conn=None, db_auto_write=False):
        super().__init__(JADE.__name__, params, db_conn, db_auto_write)

        self.archive_size = params.archive_size
        self.archive = []

        self.mutation_factor_mean = params.mutation_factor_mean
        self.mutation_factor_std = params.mutation_factor_std
        self.mutation_factors = draw_cauchy_dist_within_bounds(self.mutation_factor_mean, self.mutation_factor_std,
                                                               self.population_size)
        self.success_mutation_factors = []

        self.crossover_rate_mean = params.crossover_rate_mean
        self.crossover_rate_std = params.crossover_rate_std
        self.crossover_rate_low = params.crossover_rate_low
        self.crossover_rate_high = params.crossover_rate_high
        self.crossover_rates = draw_norm_dist_within_bounds(self.crossover_rate_mean, self.crossover_rate_std,
                                                            self.crossover_rate_low, self.crossover_rate_high,
                                                            self.population_size)
        self.success_crossover_rates = []

        self.c = params.c
        self.p = params.p

    def next_epoch(self):
        # New population after mutation
        v_pop = jade_mutation(self._pop, self.mutation_factors, self.p, self.archive)

        # Apply boundary constrains on population in place
        fix_boundary_constraints(v_pop, self.boundary_constraints_fun)

        # New population after crossing
        u_pop = jade_binomial_crossing(self._pop, v_pop, self.crossover_rates)

        # Update values before selection
        u_pop.update_fitness_values(self._function.eval, self.parallel_processing)

        # Select new population
        new_pop = jade_selection(self._pop, u_pop, self.mutation_factors, self.crossover_rates,
                                 self.success_mutation_factors, self.success_crossover_rates, self.archive)

        jade_reduce_archive(self.archive, self.archive_size)

        self.mutation_factors, self.mutation_factor_mean = jade_adapt_mutation_factors(self.c,
                                                                                       self.mutation_factor_mean,
                                                                                       self.mutation_factor_std,
                                                                                       self.population_size,
                                                                                       self.success_mutation_factors)

        self.crossover_rates, self.crossover_rate_mean = jade_adapt_crossover_rates(self.c, self.crossover_rate_mean,
                                                                                    self.crossover_rate_std,
                                                                                    self.crossover_rate_low,
                                                                                    self.crossover_rate_high,
                                                                                    self.population_size,
                                                                                    self.success_crossover_rates)

        # Override data
        self._pop = new_pop

        self._epoch_number += 1



import math
from detpy.DETAlgs.base import BaseAlg
from detpy.DETAlgs.data.alg_data import ImprovedDEData
from detpy.models.enums.boundary_constrain import fix_boundary_constraints
from detpy.DETAlgs.methods.methods_improved_de import mutation, binomial_crossing, selection


class ImprovedDE(BaseAlg):
    """
        ImprovedDE

        Links:
        https://link.springer.com/article/10.1007/s00500-023-09080-1

        References:
        Yifeng Lin · Yuer Yang · Yinyan Zhang
        Improved differential evolution with dynamic mutation parameters
        Soft Computing Optimization Published: 17 August 2023 Volume 27, pages 17923–17941, (2023)
        https://doi.org/10.1007/s00500-023-09080-1

    """

    def __init__(self, params: ImprovedDEData, db_conn=None, db_auto_write=False):
        super().__init__(ImprovedDE.__name__, params, db_conn, db_auto_write)
        self.initial_mutation_factor = params.mutation_factor  # Initial F value
        self.crossover_rate = params.crossover_rate  # Cr
        self.iteration = 0

    def dynamic_mutation_factor(self, iteration):
        """
        Implements dynamic mutation factor based on Scheme 6:
        FS(k) = 1 - 1 / (1 + exp(-iteration))
        """
        return 1 - 1 / (1 + math.exp(-iteration))

    def next_epoch(self):
        dynamic_fs = self.dynamic_mutation_factor(self.iteration)

        # New population after mutation using dynamic mutation factor
        v_pop = mutation(self._pop, fs=dynamic_fs)

        # Apply boundary constraints on population in place
        fix_boundary_constraints(v_pop, self.boundary_constraints_fun)

        # New population after crossing
        u_pop = binomial_crossing(self._pop, v_pop, cr=self.crossover_rate)

        # Update values before selection
        u_pop.update_fitness_values(self._function.eval, self.parallel_processing)

        # Select new population
        new_pop = selection(self._pop, u_pop)

        # Override population with the newly selected one
        self._pop = new_pop

        # Increment iteration count
        self.iteration += 1
        self._epoch_number += 1



import numpy as np
from detpy.DETAlgs.base import BaseAlg
from detpy.DETAlgs.data.alg_data import SADEData
from detpy.DETAlgs.methods.methods_sade import sade_mutation, sade_binomial_crossing, sade_selection
from detpy.models.enums.boundary_constrain import fix_boundary_constraints


class SADE(BaseAlg):
    """
        SADE

        Links:
        https://ieeexplore.ieee.org/abstract/document/4730987

        References:
        Wu Zhi-Feng, Huang Hou-Kuan, Yang Bei and Zhang Ying, "A modified differential evolution algorithm with
        self-adaptive control parameters," 2008 3rd International Conference on Intelligent System and Knowledge Engineering
        , Xiamen, 2008, pp. 524-527, doi: 10.1109/ISKE.2008.4730987.
    """

    def __init__(self, params: SADEData, db_conn=None, db_auto_write=False):
        super().__init__(SADE.__name__, params, db_conn, db_auto_write)

        # class specific
        self._f_arr = np.random.uniform(size=self.population_size)
        self._cr_arr = np.random.uniform(size=self.population_size)
        self.prob_f = params.prob_f
        self.prob_cr = params.prob_cr

    def next_epoch(self):
        f_arr, cr_arr, prob_f, prob_cr = (self._f_arr, self._cr_arr, self.prob_f, self.prob_cr)

        # New population after mutation
        v_pop = sade_mutation(self._pop, f_arr)

        # Apply boundary constrains on population in place
        fix_boundary_constraints(v_pop, self.boundary_constraints_fun)

        # New population after crossing
        u_pop = sade_binomial_crossing(self._pop, v_pop, cr_arr)

        # Update values before selection
        u_pop.update_fitness_values(self._function.eval, self.parallel_processing)

        # Select new population
        new_pop, f_arr, cr_arr = sade_selection(self._pop, u_pop, f_arr, cr_arr, prob_f, prob_cr)

        # Override data
        self._pop = new_pop
        self._f_arr = f_arr
        self._cr_arr = cr_arr
        self.prob_f = prob_f
        self.prob_cr = prob_cr

        self._epoch_number += 1



import copy

from enum import Enum
from detpy.models.member import Member


class MutationType(Enum):
    RAND_1 = "rand_1"
    RAND_2 = "rand_2"
    BEST_1 = "best_1"
    BEST_2 = "best_2"
    CURRENT_TO_BEST_1 = "current_to_best_1"
    RAND_TO_BEST_1 = "rand_to_best_1"


def mutation_rand_1(members: list[Member], f: float) -> Member:
    """
        Formula: v_ij = x_r1 + F(x_r2 - x_r3)
    """
    new_member = copy.deepcopy(members[0])
    new_member.chromosomes = members[0].chromosomes + (members[1].chromosomes - members[2].chromosomes) * f
    return new_member


def mutation_rand_2(members: list[Member], f: float) -> Member:
    """
        Formula: v_ij = x_r1 + F(x_r2 - x_r3) + F(x_r4 - x_r5)
    """
    new_member = copy.deepcopy(members[0])
    new_member.chromosomes = (members[0].chromosomes + (members[1].chromosomes - members[2].chromosomes) * f +
                              (members[3].chromosomes - members[4].chromosomes) * f)
    return new_member


def mutation_best_1(best_member: Member, members: list[Member], f: float) -> Member:
    """
        Formula: v_ij = x_best + F(x_r1 - x_r2)
    """
    new_member = copy.deepcopy(best_member)
    new_member.chromosomes = best_member.chromosomes + (members[0].chromosomes - members[1].chromosomes) * f
    return new_member


def mutation_best_2(best_member: Member, members: list[Member], f: float) -> Member:
    """
        Formula: v_ij = x_best + F(x_r1 - x_r2) + F(x_r3 - x_r4)
    """
    new_member = copy.deepcopy(best_member)
    new_member.chromosomes = (best_member.chromosomes + (members[0].chromosomes - members[1].chromosomes) * f +
                              (members[2].chromosomes - members[3].chromosomes) * f)
    return new_member


def mutation_curr_to_best_1(base_member: Member, best_member: Member, members: list[Member], f: float) -> Member:
    """
        Formula: v_ij = x_base + F(x_best - x_base) + F(x_r1 - x_r2)
    """
    new_member = copy.deepcopy(base_member)
    new_member.chromosomes = (base_member.chromosomes + (best_member.chromosomes - base_member.chromosomes) * f +
                              (members[0].chromosomes - members[1].chromosomes) * f)
    return new_member


def mutation_rand_to_best_1(best_member: Member, members: list[Member], f: float) -> Member:
    """
            Formula: v_ij = x_r1 + F(x_best - x_r1) + F(x_r2 - x_r3)
    """
    new_member = copy.deepcopy(members[0])
    new_member.chromosomes = (members[0].chromosomes + (best_member.chromosomes - members[0].chromosomes) * f + (
            members[1].chromosomes - members[2].chromosomes) * f)

    return new_member



import numpy as np
import math

from detpy.DETAlgs.methods.methods_de import binomial_crossing_ind
from detpy.models.population import Population


def ide_binomial_crossing(origin_population: Population, mutated_population: Population, cr_arr):
    if origin_population.size != mutated_population.size:
        print("Binomial_crossing: populations have different sizes")
        return None

    new_members = []
    for i in range(origin_population.size):
        new_member = binomial_crossing_ind(origin_population.members[i], mutated_population.members[i], cr_arr[i])
        new_members.append(new_member)

    new_population = Population(
        lb=origin_population.lb,
        ub=origin_population.ub,
        arg_num=origin_population.arg_num,
        size=origin_population.size,
        optimization=origin_population.optimization
    )
    new_population.members = np.array(new_members)
    return new_population


def ide_get_f(curr_gen, max_gen):
    f_0 = 0.5
    exponent = 1 - (max_gen / (max_gen + 1 - curr_gen))
    return f_0 * 2 ** (math.e ** exponent)


def ide_get_cr(pop: Population):
    cr_arr = np.zeros(pop.size)
    cr_min, cr_max = 0.3, 0.9

    fitness_max = pop.get_best_members(pop.size)[-1].fitness_value
    fitness_mean = pop.mean()

    for i in range(pop.size):
        member_fitness = pop.members[i].fitness_value

        if member_fitness > fitness_mean:
            cr_arr[i] = cr_min + (cr_max - cr_min) * ((member_fitness - fitness_mean) / (fitness_max - fitness_mean))
        else:
            cr_arr[i] = cr_min

    return cr_arr




import concurrent.futures
import numpy as np
from detpy.models.enums.optimization import OptimizationType
from detpy.models.member import Member


class Population:
    def __init__(self, lb, ub, arg_num, size, optimization: OptimizationType):
        self.size = size
        self.members = None
        self.optimization = optimization

        # chromosome config
        self.lb = lb
        self.ub = ub
        self.arg_num = arg_num

    def generate_population(self):
        self.members = np.array([Member(self.lb, self.ub, self.arg_num) for _ in range(self.size)])

    @staticmethod
    def calculate_fitness(member, fitness_fun):
        args = member.get_chromosomes()
        return fitness_fun(args)

    def calculate_member(self, index, fitness_fun):
        member = self.members[index]
        member.fitness_value = self.calculate_fitness(member, fitness_fun)

    def update_fitness_values(self, fitness_fun, parallel_processing=None):
        if parallel_processing is None:
            executor_class = concurrent.futures.ThreadPoolExecutor
            worker = 1
        elif parallel_processing[0] == "process":
            raise ValueError('ProcessPoolExecutor is not supported. Please use thread configuration.')
        else:
            executor_class = concurrent.futures.ThreadPoolExecutor
            worker = parallel_processing[1]

        with executor_class(max_workers=worker) as executor:
            executor.map(lambda idx: self.calculate_member(idx, fitness_fun), range(len(self.members)))

    def get_best_members(self, nr_of_members):
        # Get the indices that would sort the array based on the key function
        sorted_indices = np.argsort([member.fitness_value for member in self.members])
        # Use the sorted indices to sort the array
        sorted_array = self.members[sorted_indices]
        return sorted_array[:nr_of_members]

    def mean(self):
        return np.mean([member.fitness_value for member in self.members])

    def std(self):
        return np.std([member.fitness_value for member in self.members])

    def __str__(self, population_label=""):
        output = f"Population{population_label}:"
        for m in self.members:
            output += f"\n{str(m)}"

        return output



import numpy as np
import math
from abc import ABC, abstractmethod


class BaseChromosome(ABC):

    @abstractmethod
    def calculate_real_value(self, bin_ind):
        pass


class Chromosome(BaseChromosome):
    def __init__(self, lb, ub):
        # Options
        self.lb = lb
        self.ub = ub

        # Real value
        self.real_value = np.random.uniform(self.lb, self.ub)

    def calculate_real_value(self, bin_ind):
        binary_string = ''.join([str(elem) for elem in bin_ind])
        return self.lb + int(binary_string, 2) * (self.ub - self.lb) / (
                math.pow(2, bin_ind.size) - 1)

    def __add__(self, other):
        c = Chromosome(self.lb, self.ub)
        c.real_value = self.real_value + other.real_value
        return c

    def __sub__(self, other):
        c = Chromosome(self.lb, self.ub)
        if isinstance(other, Chromosome):
            c.real_value = self.real_value - other.real_value
        elif isinstance(other, (float, np.int32, np.int64, np.float32, np.float64)):
            c.real_value = self.real_value - other
        else:
            raise TypeError(f"Unsupported operand type(s) for -: 'Chromosome' and '{type(other).__name__}'")
        return c

    def __mul__(self, other):
        c = Chromosome(self.lb, self.ub)
        c.real_value = self.real_value * other
        return c

    def __abs__(self):
        c = Chromosome(self.lb, self.ub)
        c.real_value = abs(self.real_value)
        return c



import random
import numpy as np
import copy

from detpy.models.population import Population
from detpy.models.enums.optimization import OptimizationType
from detpy.models.enums.mutation import mutation_rand_1


def delb_mutation(population: Population):
    members = population.members.tolist()
    drew_members = [random.sample(members, 3) for _ in range(population.size)]
    drew_fs = [random.uniform(-1, -0.4) if random.random() < 0.5 else random.uniform(0.4, 1)
               for _ in range(population.size)]

    new_members = [mutation_rand_1(selected_members, f)
                   for selected_members, f in zip(drew_members, drew_fs)]

    new_population = Population(
        lb=population.lb,
        ub=population.ub,
        arg_num=population.arg_num,
        size=population.size,
        optimization=population.optimization
    )
    new_population.members = np.array(new_members)

    return new_population


def delb_selection(origin_population: Population, modified_population: Population, w: float, fitness_func):
    if origin_population.size != modified_population.size:
        print("Selection: populations have different sizes")
        return None

    if origin_population.optimization != modified_population.optimization:
        print("Selection: populations have different optimization types")
        return None

    best_member = origin_population.get_best_members(1)[0]
    optimization = origin_population.optimization

    new_members = []
    for i in range(origin_population.size):
        if optimization == OptimizationType.MINIMIZATION:
            if origin_population.members[i] <= modified_population.members[i]:
                new_members.append(copy.deepcopy(origin_population.members[i]))
            elif random.uniform(0, 1) < w and modified_population.members[i] > best_member:
                r = best_member - (modified_population.members[i] - best_member)
                r.calculate_fitness_fun(fitness_fun=fitness_func)
                if r < modified_population.members[i]:
                    new_members.append(copy.deepcopy(r))
                else:
                    c = best_member + (modified_population.members[i] - best_member) * 0.5
                    c.calculate_fitness_fun(fitness_fun=fitness_func)
                    if c < modified_population.members[i]:
                        new_members.append(copy.deepcopy(c))
                    else:
                        new_members.append(copy.deepcopy(modified_population.members[i]))
            else:
                new_members.append(copy.deepcopy(modified_population.members[i]))

        elif optimization == OptimizationType.MAXIMIZATION:
            if origin_population.members[i] >= modified_population.members[i]:
                new_members.append(copy.deepcopy(origin_population.members[i]))
            elif random.uniform(0, 1) < w and modified_population.members[i] < best_member:
                r = best_member - (modified_population.members[i] - best_member)
                r.calculate_fitness_fun(fitness_fun=fitness_func)
                if r > modified_population.members[i]:
                    new_members.append(copy.deepcopy(r))
                else:
                    c = best_member + 0.5 * (modified_population.members[i] - best_member)
                    c.calculate_fitness_fun(fitness_fun=fitness_func)
                    if c > modified_population.members[i]:
                        new_members.append(copy.deepcopy(c))
                    else:
                        new_members.append(copy.deepcopy(modified_population.members[i]))
            else:
                new_members.append(copy.deepcopy(modified_population.members[i]))

    new_population = Population(
        lb=origin_population.lb,
        ub=origin_population.ub,
        arg_num=origin_population.arg_num,
        size=origin_population.size,
        optimization=origin_population.optimization
    )
    new_population.members = np.array(new_members)

    return new_population



import os
import json
import math
import numpy as np
from scipy.stats import multivariate_normal

class Ackley:
    def __init__(self, n_dimensions):
        self.n_dimensions = n_dimensions
        self.name="Ackley"

    def evaluate_func(self, x):
        if len(x) != self.n_dimensions:
            raise ValueError(f"Ackley function requires {self.n_dimensions} variables, but {len(x)} were given.")
        a = 20
        b = 0.2
        c = 2 * math.pi
        part1 = -a * math.exp(-b * np.sqrt(np.sum(np.square(x)) / self.n_dimensions))
        part2 = -math.exp(np.sum(np.cos(c * np.array(x))) / self.n_dimensions)
        return part1 + part2 + a + math.exp(1)

class Rastrigin:
    def __init__(self, n_dimensions):
        self.n_dimensions = n_dimensions
        self.name = "Rastrigin"

    def evaluate_func(self, x):
        if len(x) != self.n_dimensions:
            raise ValueError(f"Rastrigin function requires {self.n_dimensions} variables, but {len(x)} were given.")
        A = 10
        return A * self.n_dimensions + sum([(xi ** 2 - A * np.cos(2 * np.pi * xi)) for xi in x])

class Rosenbrock:
    def __init__(self, n_dimensions):
        self.n_dimensions = n_dimensions
        self.name = "Rosenbrock"

    def evaluate_func(self, x):
        if len(x) != self.n_dimensions:
            raise ValueError(f"Rosenbrock function requires {self.n_dimensions} variables, but {len(x)} were given.")
        return sum([100 * (x[i + 1] - x[i] ** 2) ** 2 + (1 - x[i]) ** 2 for i in range(self.n_dimensions - 1)])

class Sphere:
    def __init__(self, n_dimensions):
        self.n_dimensions = n_dimensions
        self.name = "Sphere"

    def evaluate_func(self, x):
        if len(x) != self.n_dimensions:
            raise ValueError(f"Sphere function requires {self.n_dimensions} variables, but {len(x)} were given.")
        return sum([xi ** 2 for xi in x])

class Griewank:
    def __init__(self, n_dimensions):
        self.n_dimensions = n_dimensions
        self.name = "Griewank"

    def evaluate_func(self, x):
        if len(x) != self.n_dimensions:
            raise ValueError(f"Griewank function requires {self.n_dimensions} variables, but {len(x)} were given.")
        sum_part = sum([xi ** 2 / 4000 for xi in x])
        prod_part = np.prod([np.cos(xi / np.sqrt(i + 1)) for i, xi in enumerate(x)])
        return sum_part - prod_part + 1

class Schwefel:
    def __init__(self, n_dimensions):
        self.n_dimensions = n_dimensions
        self.name = "Schwefel"

    def evaluate_func(self, x):
        if len(x) != self.n_dimensions:
            raise ValueError(f"Schwefel function requires {self.n_dimensions} variables, but {len(x)} were given.")
        return 418.9829 * self.n_dimensions - sum([xi * np.sin(np.sqrt(abs(xi))) for xi in x])

class Michalewicz:
    def __init__(self, n_dimensions, m=10):
        self.n_dimensions = n_dimensions
        self.m = m
        self.name = "Michalewicz"

    def evaluate_func(self, x):
        if len(x) != self.n_dimensions:
            raise ValueError(f"Michalewicz function requires {self.n_dimensions} variables, but {len(x)} were given.")
        return -sum([math.sin(xi) * (math.sin((i + 1) * xi ** 2 / math.pi) ** (2 * self.m)) for i, xi in enumerate(x)])

class Easom:
    def __init__(self):
        self.n_dimensions = 2
        self.name = "Easom"

    def evaluate_func(self, x):
        if len(x) != self.n_dimensions:
            raise ValueError(f"Easom function requires 2 variables, but {len(x)} were given.")
        return -math.cos(x[0]) * math.cos(x[1]) * math.exp(-((x[0] - math.pi) ** 2 + (x[1] - math.pi) ** 2))

class Himmelblau:
    def __init__(self):
        self.n_dimensions = 2
        self.name = "Himmelblau"

    def evaluate_func(self, x):
        if len(x) != self.n_dimensions:
            raise ValueError(f"Himmelblau function requires 2 variables, but {len(x)} were given.")
        return (x[0] ** 2 + x[1] - 11) ** 2 + (x[0] + x[1] ** 2 - 7) ** 2

class Keane:
    def __init__(self, n_dimensions):
        self.n_dimensions = n_dimensions
        self.name = "Keane"

    def evaluate_func(self, x):
        if len(x) != self.n_dimensions:
            raise ValueError(f"Keane function requires {self.n_dimensions} variables, but {len(x)} were given.")
        part0 = np.prod([np.cos(xi) ** 2 for xi in x])
        part1 = abs(sum([np.cos(xi) ** 4 for xi in x]) - 2 * part0)
        part2 = math.sqrt(sum([(i + 1) * xi ** 2 for i, xi in enumerate(x)]))
        return -part1 / part2

class Rana:
    def __init__(self, n_dimensions):
        self.n_dimensions = n_dimensions
        self.name = "Rana"

    def evaluate_func(self, x):
        if len(x) != self.n_dimensions:
            raise ValueError(f"Rana function requires {self.n_dimensions} variables, but {len(x)} were given.")
        s = 0.0
        for i in range(self.n_dimensions - 1):
            s += x[i] * math.cos(math.sqrt(abs(x[i + 1] + x[i] + 1))) * math.sin(math.sqrt(abs(x[i + 1] - x[i] + 1)))
        return s

class PitsAndHoles:
    def __init__(self):
        self.n_dimensions = 2
        self.mu = [[0, 0], [20, 0], [0, 20], [-20, 0], [0, -20], [10, 10], [-10, -10], [-10, 10], [10, -10]]
        self.c = [10.5, 14.0, 16.0, 12.0, 9.0, 0.1, 0.2, 0.25, 0.17]
        self.v = [2.0, 2.5, 2.7, 2.5, 2.3, 0.05, 0.3, 0.24, 0.23]
        self.name = "PitsAndHoles"

    def _get_covariance_matrix(self, idx):
        return [[self.c[idx], 0], [0, self.c[idx]]]

    def evaluate_func(self, x):
        if len(x) != self.n_dimensions:
            raise ValueError(f"Pits and Holes function requires 2 variables, but {len(x)} were given.")
        v = 0
        for i in range(len(self.mu)):
            v += multivariate_normal.pdf(x, mean=self.mu[i], cov=self._get_covariance_matrix(i)) * self.v[i]
        return -v

class Hypersphere:
    def __init__(self, n_dimensions):
        self.n_dimensions = n_dimensions
        self.name = "Hypersphere"

    def evaluate_func(self, x):
        if len(x) != self.n_dimensions:
            raise ValueError(f"Hypersphere function requires {self.n_dimensions} variables, but {len(x)} were given.")
        return sum([xi ** 2 for xi in x])

class Hyperellipsoid:
    def __init__(self, n_dimensions):
        self.n_dimensions = n_dimensions
        self.name = "Hyperellipsoid"

    def evaluate_func(self, x):
        if len(x) != self.n_dimensions:
            raise ValueError(f"Hyperellipsoid function requires {self.n_dimensions} variables, but {len(x)} were given.")
        return sum([sum([xj ** 2 for xj in x[:i + 1]]) for i in range(self.n_dimensions)])

class EggHolder:
    def __init__(self, n_dimensions):
        self.n_dimensions = n_dimensions
        self.name = "EggHolder"

    def evaluate_func(self, x):
        if len(x) != self.n_dimensions:
            raise ValueError(f"Egg Holder function requires {self.n_dimensions} variables, but {len(x)} were given.")
        return sum([(x[i + 1] + 47) * math.sin(math.sqrt(abs(x[i + 1] + 47 + x[i] / 2))) + x[i] * math.sin(math.sqrt(abs(x[i] - (x[i + 1] + 47)))) for i in range(self.n_dimensions - 1)])

class StyblinskiTang:
    def __init__(self, n_dimensions):
        self.n_dimensions = n_dimensions
        self.name = "StyblinskiTang"

    def evaluate_func(self, x):
        if len(x) != self.n_dimensions:
            raise ValueError(f"Styblinski-Tang function requires {self.n_dimensions} variables, but {len(x)} were given.")
        return sum([xi ** 4 - 16 * xi ** 2 + 5 * xi for xi in x]) / 2

class GoldsteinAndPrice:
    def __init__(self):
        self.n_dimensions = 2
        self.name = "GoldsteinAndPrice"

    def evaluate_func(self, x):
        if len(x) != self.n_dimensions:
            raise ValueError(f"Goldstein and Price function requires 2 variables, but {len(x)} were given.")
        part1 = (1 + (x[0] + x[1] + 1) ** 2 * (19 - 14 * x[0] + 3 * x[0] ** 2 - 14 * x[1] + 6 * x[0] * x[1] + 3 * x[1] ** 2))
        part2 = (30 + (2 * x[0] - 3 * x[1]) ** 2 * (18 - 32 * x[0] + 12 * x[0] ** 2 + 48 * x[1] - 36 * x[0] * x[1] + 27 * x[1] ** 2))
        return part1 * part2

class FunctionLoader:
    def __init__(self):
        self.folder_path = 'functions_info'
        self.functions = self.load_all_functions()
        self.function_classes = {
            "ackley": Ackley,
            "rastrigin": Rastrigin,
            "rosenbrock": Rosenbrock,
            "sphere": Sphere,
            "griewank": Griewank,
            "schwefel": Schwefel,
            "michalewicz": Michalewicz,
            "easom": Easom,
            "himmelblau": Himmelblau,
            "keane": Keane,
            "rana": Rana,
            "pits_and_holes": PitsAndHoles,
            "hypersphere": Hypersphere,
            "hyperellipsoid": Hyperellipsoid,
            "eggholder": EggHolder,
            "styblinski_tang": StyblinskiTang,
            "goldstein_and_price": GoldsteinAndPrice
        }

    def load_all_functions(self):
        functions = {}
        for file_name in os.listdir(os.path.join(os.path.dirname(os.path.realpath(__file__)), self.folder_path)):
            if file_name.endswith('.json'):
                function_name = file_name.replace('.json', '')
                functions[function_name] = self.load_function_from_json(file_name)
        return functions

    def load_function_from_json(self, file_name):
        file_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), self.folder_path, file_name)
        if not os.path.exists(file_path):
            raise FileNotFoundError(f"File '{file_path}' does not exist.")
        with open(file_path, 'r') as file:
            function_data = json.load(file)
        return function_data

    def get_function(self, function_name, n_dimensions):
        if function_name in ["himmelblau", "easom", "pits_and_holes", "goldstein_and_price"]:
            return self.function_classes[function_name]()
        elif function_name in self.function_classes:
            return self.function_classes[function_name](n_dimensions)
        else:
            raise ValueError(f"Function '{function_name}' not found.")

    def evaluate_function(self, function_name, variables, n_dimensions=None):
        if n_dimensions is None:
            n_dimensions = len(variables)
        function_instance = self.get_function(function_name, n_dimensions)
        if len(variables) != n_dimensions:
            raise ValueError(f"Function '{function_name}' requires {n_dimensions} variables, but {len(variables)} were given.")
        return function_instance.evaluate_func(variables)


import setuptools
import os
import io

current_path = os.path.dirname(os.path.realpath(__file__))

with io.open(f"{current_path}/README.md", mode="r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="detpy",
    packages=setuptools.find_packages(),
    package_data={
        "detpy": ["functions/functions_info/*.json"],
    },
    version="1.0.10",
    author="Szymon Ściegienny, Błażej Zieliński, Hubert Orlicki, Wojciech Książek",
    author_email="wojciech.ksiazek@pk.edu.pl",
    description="DetPy (Differential Evolution Tools): A Python toolbox for solving optimization problems "
                "using differential evolution",
    keywords=["optimization", "metaheuristics", "nature-inspired algorithms",
              "evolutionary computation", "population-based algorithms",
              "Stochastic optimization", "different evolution"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Blazej-Zielinski/detpy",
    include_package_data=True,
    license="GPLv3",
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: System :: Benchmark",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
    ],
    install_requires=["numpy", "opfunu", "matplotlib", "tqdm", "scipy"],
    python_requires='>=3.7',
)



import random
import numpy as np

from detpy.DETAlgs.methods.methods_de import mutation_ind
from detpy.models.population import Population


def derl_mutation(population: Population):
    new_members = []
    for _ in range(population.size):
        selected_members = np.array(random.sample(population.members.tolist(), 3))
        sorted_indices = np.argsort([member.fitness_value for member in selected_members])
        best_member, better_member, worst_member = selected_members[sorted_indices]

        f = random.uniform(-0.6, -0.4) if random.random() < 0.5 else random.uniform(0.4, 0.6)  # (−1, -0.4) ∪ (0.4, 1)
        new_member = mutation_ind(best_member, better_member, worst_member, f)

        new_members.append(new_member)

    new_population = Population(
        lb=population.lb,
        ub=population.ub,
        arg_num=population.arg_num,
        size=population.size,
        optimization=population.optimization
    )
    new_population.members = np.array(new_members)
    return new_population



from detpy.DETAlgs.base import BaseAlg
from detpy.DETAlgs.data.alg_data import AADEData
from detpy.DETAlgs.methods.methods_aade import aade_mutation, aade_crossing, aade_selection, \
    aade_adapat_parameters
from detpy.models.enums.boundary_constrain import fix_boundary_constraints


class AADE(BaseAlg):
    """
        AADE

        Links:
        https://ieeexplore.ieee.org/document/8819749

        References:
        V. Sharma, S. Agarwal and P. K. Verma, "Auto Adaptive Differential Evolution Algorithm,"
        2019 3rd International Conference on Computing Methodologies and Communication (ICCMC),
        Erode, India, 2019, pp. 958-963, doi: 10.1109/ICCMC.2019.8819749.
    """

    def __init__(self, params: AADEData, db_conn=None, db_auto_write=False):
        super().__init__(AADE.__name__, params, db_conn, db_auto_write)

        self.mutation_factors = [[params.mutation_factor, False] for _ in range(params.population_size)]
        self.crossover_rates = [[params.crossover_rate, False] for _ in range(params.population_size)]

    def next_epoch(self):
        # New population after mutation
        v_pop = aade_mutation(self._pop, mutation_factors=self.mutation_factors)

        # Apply boundary constrains on population in place
        fix_boundary_constraints(v_pop, self.boundary_constraints_fun)

        # New population after crossing
        u_pop = aade_crossing(self._pop, v_pop, crossover_rates=self.crossover_rates)

        # Update values before selection
        u_pop.update_fitness_values(self._function.eval, self.parallel_processing)

        # Select new population
        new_pop = aade_selection(self._pop, u_pop, self.mutation_factors, self.crossover_rates)

        aade_adapat_parameters(self._pop, self.mutation_factors, self.crossover_rates)

        # Override data
        self._pop = new_pop

        self._epoch_number += 1



from detpy.DETAlgs.base import BaseAlg
from detpy.DETAlgs.data.alg_data import MGDEData
from detpy.DETAlgs.methods.methods_de import binomial_crossing, selection
from detpy.DETAlgs.methods.methods_mgde import mgde_mutation, mgde_adapt_threshold
from detpy.models.enums.boundary_constrain import fix_boundary_constraints


class MGDE(BaseAlg):
    """
        MGDE

        Links:
        https://link.springer.com/article/10.1007/s10479-022-04641-3

        References:
        Zouache, D., Ben Abdelaziz, F. MGDE: a many-objective guided differential evolution with strengthened dominance
        relation and bi-goal evolution. Ann Oper Res (2022). https://doi.org/10.1007/s10479-022-04641-3
    """

    def __init__(self, params: MGDEData, db_conn=None, db_auto_write=False):
        super().__init__(MGDE.__name__, params, db_conn, db_auto_write)

        self.mutation_factor_f = params.mutation_factor_f
        self.mutation_factor_k = params.mutation_factor_k
        self.crossover_rate = params.crossover_rate
        self.threshold = params.threshold
        self.mu = params.mu
        self.generation = 1

    def next_epoch(self):
        # New population after mutation
        v_pop = mgde_mutation(self._pop, self.generation, self.num_of_epochs, self.mutation_factor_f,
                              self.mutation_factor_k)

        # Apply boundary constrains on population in place
        fix_boundary_constraints(v_pop, self.boundary_constraints_fun)

        # New population after crossing
        u_pop = binomial_crossing(self._pop, v_pop, cr=self.crossover_rate)

        # Update values before selection
        u_pop.update_fitness_values(self._function.eval, self.parallel_processing)

        # Select new population
        new_pop = selection(self._pop, u_pop)

        mgde_adapt_threshold(new_pop, self.threshold, self.mu, self._function.eval)

        # Override data
        self._pop = new_pop

        self._epoch_number += 1



from matplotlib import pyplot as plt


class AlgorithmResult:
    def __init__(self, epoch_metrics, avg_fitness, std_fitness, best_solution):
        self.epoch_metrics = epoch_metrics
        self.avg_fitness = avg_fitness
        self.std_fitness = std_fitness
        self.best_solution = best_solution

    def __repr__(self):
        return (f"AlgorithmResult(avg_fitness={self.avg_fitness}, std_fitness={self.std_fitness}, "
                f"best_solution={self.best_solution})")

    def plot_results(self, best_fitness_values, avg_fitness_values, std_fitness_values, num_of_epochs, method_name="Method"):
        epochs = range(1, num_of_epochs + 1)

        plt.figure()
        plt.plot(epochs, best_fitness_values, label="Best Fitness")
        plt.xlabel('Epoch')
        plt.ylabel('Best Fitness Value')
        plt.title(f'Best Fitness per Epoch - {method_name}')
        plt.legend()
        plt.show()

        plt.figure()
        plt.plot(epochs, avg_fitness_values, label="Average Fitness", color="orange")
        plt.xlabel('Epoch')
        plt.ylabel('Average Fitness Value')
        plt.title(f'Average Fitness per Epoch - {method_name}')
        plt.legend()
        plt.show()

        plt.figure()
        plt.plot(epochs, std_fitness_values, label="Standard Deviation of Fitness", color="green")
        plt.xlabel('Epoch')
        plt.ylabel('Standard Deviation')
        plt.title(f'Standard Deviation of Fitness per Epoch - {method_name}')
        plt.legend()
        plt.show()



import numpy as np
from detpy.models.chromosome import Chromosome


class Member:
    def __init__(self, lb, ub, args_num):

        if len(lb) < args_num or len(ub) < args_num:
            raise ValueError(
                f"Insufficient bounds for {args_num} dimensions: "
                f"received {len(lb)} lower bounds and {len(ub)} upper bounds. "
                "Ensure lb and ub lists match the dimension count."
            )

        self.chromosomes = np.array([Chromosome(lb[i], ub[i]) for i in range(args_num)])
        self.fitness_value = None
        self.lb = lb
        self.ub = ub
        self.args_num = args_num

    def calculate_fitness_fun(self, fitness_fun):
        self.fitness_value = fitness_fun([chromosome.real_value for chromosome in self.chromosomes])

    def get_chromosomes(self):
        return [chromosome.real_value for chromosome in self.chromosomes]

    def is_member_in_interval(self):
        for i in range(len(self.chromosomes)):
            if not (self.lb[i] <= self.chromosomes[i].real_value <= self.ub[1]):
                return False
        return True

    def __str__(self):
        return f"Member: [\n" \
               f"\t Real values [" \
               f"{''.join(str(chromosome.real_value) + '; ' for chromosome in self.chromosomes)}] \n" \
               f"\t Fitness value: {self.fitness_value}\n" \
               f"]"

    def __add__(self, other):
        chromosomes = self.chromosomes + other.chromosomes
        new_member = Member(self.lb, self.ub, self.args_num)
        new_member.chromosomes = chromosomes
        return new_member

    def __sub__(self, other):
        chromosomes = self.chromosomes - other.chromosomes
        new_member = Member(self.lb, self.ub, self.args_num)
        new_member.chromosomes = chromosomes
        return new_member

    def __mul__(self, other):
        chromosomes = self.chromosomes * other
        new_member = Member(self.lb, self.ub, self.args_num)
        new_member.chromosomes = chromosomes
        return new_member

    def __lt__(self, other):
        return self.fitness_value < other.fitness_value

    def __eq__(self, other):
        return self.fitness_value == other.fitness_value

    def __gt__(self, other):
        return self.fitness_value > other.fitness_value

    def __le__(self, other):
        return self.fitness_value <= other.fitness_value

    def __ne__(self, other):
        return self.fitness_value != other.fitness_value

    def __ge__(self, other):
        return self.fitness_value >= other.fitness_value

    def __abs__(self):
        # Implementation of the __abs__() method
        abs_chromosomes = abs(self.chromosomes)
        new_member = Member(self.lb, self.ub, self.args_num)
        new_member.chromosomes = abs_chromosomes
        return new_member
